#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from pyowm.commons import exceptions
from pyowm.utils import formatting


class AggregatedMeasurement:
    """
    A class representing an aggregation of measurements done by the Stations API
    on a specific time-frame. Values for the aggregation time-frame can be: 'm'
    (minute), 'h' (hour) or 'd' (day)

    :param station_id: unique station identifier
    :type station_id: str
    :param timestamp: reference UNIX timestamp for this measurement
    :type timestamp: int
    :param aggregated_on: aggregation time-frame for this measurement
    :type aggregated_on: string between 'm','h' and 'd'
    :param temp: optional dict containing temperature data
    :type temp: dict or `None`
    :param humidity: optional dict containing humidity data
    :type humidity: dict or `None`
    :param wind: optional dict containing wind data
    :type wind: dict or `None`
    :param pressure: optional dict containing pressure data
    :type pressure: dict or `None`
    :param precipitation: optional dict containing precipitation data
    :type precipitation: dict or `None`
    """

    ALLOWED_AGGREGATION_TIME_FRAMES = ['m', 'h', 'd']

    def __init__(self, station_id, timestamp, aggregated_on, temp=None,
                 humidity=None, wind=None, pressure=None, precipitation=None):
        assert station_id is not None
        assert isinstance(timestamp, int)
        assert not timestamp < 0
        assert aggregated_on is not None
        if aggregated_on not in self.ALLOWED_AGGREGATION_TIME_FRAMES:
            raise ValueError('"aggregated_on" must be among: m, h, d')
        self.station_id = station_id
        self.timestamp = timestamp
        self.aggregated_on = aggregated_on
        self.temp = dict() if temp is None else temp
        self.humidity = dict() if humidity is None else humidity
        self.wind = dict() if wind is None else wind
        self.pressure = dict() if pressure is None else pressure
        self.precipitation = dict() if precipitation is None else precipitation

    def creation_time(self, timeformat='unix'):
        """Returns the UTC time of creation of this aggregated measurement

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time, '*iso*' for ISO8601-formatted
            string in the format ``YYYY-MM-DD HH:MM:SS+00`` or `date` for
            a ``datetime.datetime`` object
        :type timeformat: str
        :returns: an int or a str or a ``datetime.datetime`` object or None
        :raises: ValueError

        """
        if self.timestamp is None:
            return None
        return formatting.timeformat(self.timestamp, timeformat)

    @classmethod
    def from_dict(cls, the_dict):
        """
        Parses an *AggregatedMeasurement* instance out of a data dictionary. Only certain properties of the data dictionary
        are used: if these properties are not found or cannot be parsed, an exception is issued.

        :param the_dict: the input dictionary
        :type the_dict: `dict`
        :returns: an *AggregatedMeasurement* instance or ``None`` if no data is available
        :raises: *ParseAPIResponseError* if it is impossible to find or parse the data needed to build the result

        """
        if the_dict is None:
            raise exceptions.ParseAPIResponseError('Data is None')
        station_id = the_dict.get('station_id', None)
        ts = the_dict.get('date', None)
        if ts is not None:
            ts = int(ts)
        aggregated_on = the_dict.get('type', None)
        temp = the_dict.get('temp', dict())
        humidity = the_dict.get('humidity', dict())
        wind = the_dict.get('wind', dict())
        pressure = the_dict.get('pressure', dict())
        precipitation = the_dict.get('precipitation', dict())
        return AggregatedMeasurement(station_id, ts, aggregated_on, temp=temp, humidity=humidity, wind=wind,
                                     pressure=pressure, precipitation=precipitation)

    def to_dict(self):
        """Dumps object fields into a dict

        :returns: a dict

        """
        return {'station_id': self.station_id,
                'timestamp': self.timestamp,
                'aggregated_on': self.aggregated_on,
                'temp': self.temp,
                'humidity': self.humidity,
                'wind': self.wind,
                'pressure': self.pressure,
                'precipitation': self.precipitation}

    def __repr__(self):
        return '<%s.%s - station_id=%s, created_at=%s>' \
               % (__name__, self.__class__.__name__,
                  self.station_id, self.creation_time())


class Measurement:

    def __init__(self, station_id, timestamp, temperature=None, wind_speed=None,
                 wind_gust=None, wind_deg=None, pressure=None, humidity=None,
                 rain_1h=None, rain_6h=None, rain_24h=None, snow_1h=None,
                 snow_6h=None, snow_24h=None, dew_point=None, humidex=None,
                 heat_index=None, visibility_distance=None, visibility_prefix=None,
                 clouds_distance=None, clouds_condition=None, clouds_cumulus=None,
                 weather_precipitation=None, weather_descriptor=None,
                 weather_intensity=None, weather_proximity=None,
                 weather_obscuration=None, weather_other=None):
        assert station_id is not None
        assert isinstance(timestamp, int)
        assert not timestamp < 0
        self.station_id = station_id
        self.timestamp = timestamp
        self.temperature = temperature
        self.wind_speed = wind_speed
        self.wind_gust = wind_gust
        self.wind_deg = wind_deg
        self.pressure = pressure
        self.humidity = humidity
        self.rain_1h = rain_1h
        self.rain_6h = rain_6h
        self.rain_24h = rain_24h
        self.snow_1h = snow_1h
        self.snow_6h = snow_6h
        self.snow_24h = snow_24h
        self.dew_point = dew_point
        self.humidex = humidex
        self.heat_index = heat_index
        self.visibility_distance = visibility_distance
        self.visibility_prefix = visibility_prefix
        self.clouds_distance = clouds_distance
        self.clouds_condition = clouds_condition
        self.clouds_cumulus = clouds_cumulus
        self.weather_precipitation = weather_precipitation
        self.weather_descriptor = weather_descriptor
        self.weather_intensity = weather_intensity
        self.weather_proximity = weather_proximity
        self.weather_obscuration = weather_obscuration
        self.weather_other = weather_other

    def creation_time(self, timeformat='unix'):
        """Returns the UTC time of creation of this raw measurement

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time, '*iso*' for ISO8601-formatted
            string in the format ``YYYY-MM-DD HH:MM:SS+00`` or `date` for
            a ``datetime.datetime`` object
        :type timeformat: str
        :returns: an int or a str or a ``datetime.datetime`` object or None
        :raises: ValueError

        """
        if self.timestamp is None:
            return None
        return formatting.timeformat(self.timestamp, timeformat)

    @classmethod
    def from_dict(cls, the_dict):
        if 'station_id' not in the_dict:
            raise KeyError('"station_id" must be provided')
        station_id = the_dict['station_id']
        if 'timestamp' not in the_dict:
            raise KeyError('"timestamp" must be provided')
        timestamp = the_dict['timestamp']

        temperature = the_dict.get('temperature', None)
        wind_speed = the_dict.get('wind_speed', None)
        wind_gust = the_dict.get('wind_gust', None)
        wind_deg = the_dict.get('wind_deg', None)
        pressure = the_dict.get('pressure', None)
        humidity = the_dict.get('humidity', None)
        rain_1h = the_dict.get('rain_1h', None)
        rain_6h = the_dict.get('rain_6h', None)
        rain_24h = the_dict.get('rain_24h', None)
        snow_1h = the_dict.get('snow_1h', None)
        snow_6h = the_dict.get('snow_6h', None)
        snow_24h = the_dict.get('snow_24h', None)
        dew_point = the_dict.get('dew_point', None)
        humidex = the_dict.get('humidex', None)
        heat_index = the_dict.get('heat_index', None)
        visibility_distance = the_dict.get('visibility_distance', None)
        visibility_prefix = the_dict.get('visibility_prefix', None)
        clouds_distance = the_dict.get('clouds_distance', None)
        clouds_condition = the_dict.get('clouds_condition', None)
        clouds_cumulus = the_dict.get('clouds_cumulus', None)
        weather_precipitation = the_dict.get('weather_precipitation', None)
        weather_descriptor = the_dict.get('weather_descriptor', None)
        weather_intensity = the_dict.get('weather_intensity', None)
        weather_proximity = the_dict.get('weather_proximity', None)
        weather_obscuration = the_dict.get('weather_obscuration', None)
        weather_other = the_dict.get('weather_other', None)
        return Measurement(station_id, timestamp, temperature=temperature,
            wind_speed=wind_speed, wind_gust=wind_gust, wind_deg=wind_deg,
            pressure=pressure, humidity=humidity,rain_1h=rain_1h,rain_6h=rain_6h,
            rain_24h=rain_24h,snow_1h=snow_1h,snow_6h=snow_6h,snow_24h=snow_24h,
            dew_point=dew_point,humidex=humidex,heat_index=heat_index,
            visibility_distance=visibility_distance,visibility_prefix=visibility_prefix,
            clouds_distance=clouds_distance,clouds_condition=clouds_condition,
            clouds_cumulus=clouds_cumulus,weather_precipitation=weather_precipitation,
            weather_descriptor=weather_descriptor,weather_intensity=weather_intensity,
            weather_proximity=weather_proximity,weather_obscuration=weather_obscuration,
            weather_other=weather_other)

    def to_dict(self):
        """Dumps object fields into a dictionary

        :returns: a dict

        """
        return {
            'station_id': self.station_id,
            'timestamp': self.timestamp,
            'temperature': self.temperature,
            'wind_speed': self.wind_speed,
            'wind_gust': self.wind_gust,
            'wind_deg': self.wind_deg,
            'pressure': self.pressure,
            'humidity': self.humidity,
            'rain_1h': self.rain_1h,
            'rain_6h': self.rain_6h,
            'rain_24h': self.rain_24h,
            'snow_1h': self.snow_1h,
            'snow_6h': self.snow_6h,
            'snow_24h': self.snow_24h,
            'dew_point': self.dew_point,
            'humidex': self.humidex,
            'heat_index': self.heat_index,
            'visibility_distance': self.visibility_distance,
            'visibility_prefix': self.visibility_prefix,
            'clouds_distance': self.clouds_distance,
            'clouds_condition': self.clouds_condition,
            'clouds_cumulus': self.clouds_cumulus,
            'weather_precipitation': self.weather_precipitation,
            'weather_descriptor': self.weather_descriptor,
            'weather_intensity': self.weather_intensity,
            'weather_proximity': self.weather_proximity,
            'weather_obscuration': self.weather_obscuration,
            'weather_other': self.weather_other}

    def to_JSON(self):
        """Dumps object fields into a JSON formatted string

        :returns: the JSON string

        """
        return json.dumps(self.to_dict())

    def __repr__(self):
        return '<%s.%s - station_id=%s, created_at=%s>' \
               % (__name__, self.__class__.__name__,
                  self.station_id, self.creation_time())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.utils.geo import GeometryBuilder
from pyowm.utils.geo import Point as GeoPoint
from pyowm.utils.geo import Polygon as GeoPolygon


class Polygon:

    """
    A Polygon feature, foundational element for all Agro API operations

    :param id: the polygon's ID
    :type id: str
    :param name: the polygon's name
    :type namr: str
    :param geopolygon: the `pyowm.utils.geo.Polygon` instance that represents this polygon
    :type geopolygon: `pyowm.utils.geo.Polygon`
    :param center: the `pyowm.utils.geo.Point` instance that represents the central point of the polygon
    :type center: `pyowm.utils.geo.Point`
    :param area: the area of the polygon in hectares
    :type area: float or int
    :param user_id: the ID of the user owning this polygon
    :type user_id: str
    :returns: a `Polygon` instance
    :raises: `AssertionError` when either id is `None` or geopolygon, center or area have wrong type
    """

    def __init__(self, id, name=None, geopolygon=None, center=None, area=None, user_id=None):

        assert id is not None, 'Polygon ID cannot be None'
        if geopolygon is not None:
            assert isinstance(geopolygon, GeoPolygon), 'Polygon must be a valid geopolygon type'
        if center is not None:
            assert isinstance(center, GeoPoint), 'Polygon center must be a valid geopoint type'
        if area is not None:
            assert isinstance(area, (float, int)), 'Area must be a numeric type'
            assert area >= 0, 'Area must not be negative'
        self.id = id
        self.name = name
        self.geopolygon = geopolygon
        self.center = center
        self.area = area
        self.user_id = user_id

    @property
    def area_km(self):
        if self.area:
            return self.area * 0.01
        return None

    @classmethod
    def from_dict(cls, the_dict):
        assert isinstance(the_dict, dict)
        the_id = the_dict.get('id', None)
        geojson = the_dict.get('geo_json', {}).get('geometry', None)
        name = the_dict.get('name', None)
        center = the_dict.get('center', None)
        area = the_dict.get('area', None)
        user_id =the_dict.get('user_id', None)
        geopolygon = GeometryBuilder.build(geojson)
        try:
            center = GeoPoint(center[0], center[1])
        except:
            raise ValueError('Wrong format for polygon center coordinates')
        return Polygon(the_id, name, geopolygon, center, area, user_id)

    def __repr__(self):
        return "<%s.%s - id=%s, name=%s, area=%s>" % (__name__, self.__class__.__name__, self.id, self.name, str(self.area))



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class MapLayerEnum:
    """
    Allowed map layer values for tiles retrieval

    """
    PRECIPITATION = 'precipitation_new'
    WIND = 'wind_new'
    TEMPERATURE = 'temp_new'
    PRESSURE = 'pressure_new'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.commons import exceptions


def status_is(weather, status, weather_code_registry):
    """
    Checks if the weather status code of a *Weather* object corresponds to the
    detailed status indicated. The lookup is performed against the provided 
    *WeatherCodeRegistry* object.

    :param weather: the *Weather* object whose status code is to be checked
    :type weather: *Weather*
    :param status: a string indicating a detailed weather status
    :type status: str
    :param weather_code_registry: a *WeatherCodeRegistry* object
    :type weather_code_registry: *WeatherCodeRegistry*
    :returns: ``True`` if the check is positive, ``False`` otherwise

    """
    weather_status = weather_code_registry. \
        status_for(weather.weather_code).lower()
    return weather_status == status


def any_status_is(weather_list, status, weather_code_registry):
    """
    Checks if the weather status code of any of the *Weather* objects in the
    provided list corresponds to the detailed status indicated. The lookup is
    performed against the provided *WeatherCodeRegistry* object.

    :param weathers: a list of *Weather* objects
    :type weathers: list
    :param status: a string indicating a detailed weather status
    :type status: str
    :param weather_code_registry: a *WeatherCodeRegistry* object
    :type weather_code_registry: *WeatherCodeRegistry*
    :returns: ``True`` if the check is positive, ``False`` otherwise
    
    """
    return any(
        status_is(weather, status, weather_code_registry)
        for weather in weather_list
    )


def filter_by_status(weather_list, status, weather_code_registry):
    """
    Filters out from the provided list of *Weather* objects a sublist of items
    having a status corresponding to the provided one. The lookup is performed
    against the provided *WeatherCodeRegistry* object.

    :param weathers: a list of *Weather* objects
    :type weathers: list
    :param status: a string indicating a detailed weather status
    :type status: str
    :param weather_code_registry: a *WeatherCodeRegistry* object
    :type weather_code_registry: *WeatherCodeRegistry*
    :returns: ``True`` if the check is positive, ``False`` otherwise
    
    """
    return [
        weather
        for weather in weather_list
        if status_is(weather, status, weather_code_registry)
    ]


def is_in_coverage(unixtime, weathers_list):
    """
    Checks if the supplied UNIX time is contained into the time range
    (coverage) defined by the most ancient and most recent *Weather* objects
    in the supplied list

    :param unixtime: the UNIX time to be searched in the time range
    :type unixtime: int
    :param weathers_list: the list of *Weather* objects to be scanned for
        global time coverage
    :type weathers_list: list
    :returns: ``True`` if the UNIX time is contained into the time range,
        ``False`` otherwise
    """
    if not weathers_list:
        return False
    min_of_coverage = min(weather.reference_time() for weather in weathers_list)
    max_of_coverage = max([weather.reference_time() \
                           for weather in weathers_list])
    return unixtime >= min_of_coverage and unixtime <= max_of_coverage


def find_closest_weather(weathers_list, unixtime):
    """
    Extracts from the provided list of Weather objects the item which is
    closest in time to the provided UNIXtime.

    :param weathers_list: a list of *Weather* objects
    :type weathers_list: list
    :param unixtime: a UNIX time
    :type unixtime: int
    :returns: the *Weather* object which is closest in time or ``None`` if the
        list is empty
    """
    if not weathers_list:
        return None
    if not is_in_coverage(unixtime, weathers_list):
        raise exceptions.NotFoundError('Error: the specified time is ' + \
                                'not included in the weather coverage range')
    closest_weather = weathers_list[0]
    time_distance = abs(closest_weather.reference_time() - unixtime)
    for weather in weathers_list:
        if abs(weather.reference_time() - unixtime) < time_distance:
            time_distance = abs(weather.reference_time() - unixtime)
            closest_weather = weather
    return closest_weather



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.alertapi30.alert import Alert
from pyowm.alertapi30.trigger import Trigger
from pyowm.alertapi30.uris import ROOT_ALERT_API_URL, TRIGGERS_URI, NAMED_TRIGGER_URI, ALERTS_URI, NAMED_ALERT_URI
from pyowm.commons.http_client import HttpClient
from pyowm.constants import ALERT_API_VERSION
from pyowm.utils import formatting, timestamps


class AlertManager:

    """
    A manager objects that provides a full interface to OWM Alert API. It implements CRUD methods on Trigger entities
    and read/deletion of related Alert objects

    :param API_key: the OWM Weather API key
    :type API_key: str
    :param config: the configuration dictionary
    :type config: dict
    :returns: an *AlertManager* instance
    :raises: *AssertionError* when no API Key is provided

    """

    def __init__(self, API_key, config):
        assert API_key is not None, 'You must provide a valid API Key'
        self.API_key = API_key
        assert isinstance(config, dict)
        self.http_client = HttpClient(API_key, config, ROOT_ALERT_API_URL)

    def alert_api_version(self):
        return ALERT_API_VERSION

    # TRIGGER methods

    def create_trigger(self,  start, end, conditions, area, alert_channels=None):
        """
        Create a new trigger on the Alert API with the given parameters
        :param start: time object representing the time when the trigger begins to be checked
        :type start: int, ``datetime.datetime`` or ISO8601-formatted string
        :param end: time object representing the time when the trigger ends to be checked
        :type end: int, ``datetime.datetime`` or ISO8601-formatted string
        :param conditions: the `Condition` objects representing the set of checks to be done on weather variables
        :type conditions: list of `pyowm.utils.alertapi30.Condition` instances
        :param area: the geographic are over which conditions are checked: it can be composed by multiple geoJSON types
        :type area: list of geoJSON types
        :param alert_channels: the alert channels through which alerts originating from this `Trigger` can be consumed.
        Defaults to OWM API polling
        :type alert_channels: list of `pyowm.utils.alertapi30.AlertChannel` instances
        :returns:  a *Trigger* instance
        :raises: *ValueError* when start or end epochs are `None` or when end precedes start or when conditions or area
        are empty collections
        """
        assert start is not None
        assert end is not None

        # prepare time period
        unix_start = formatting.to_UNIXtime(start)
        unix_end = formatting.to_UNIXtime(end)
        unix_current = timestamps.now(timeformat='unix')
        if unix_start >= unix_end:
            raise ValueError("The start timestamp must precede the end timestamp")
        delta_millis_start = timestamps.millis_offset_between_epochs(unix_current, unix_start)
        delta_millis_end = timestamps.millis_offset_between_epochs(unix_current, unix_end)
        the_time_period = {
            "start": {
                "expression": "after",
                "amount": delta_millis_start
            },
            "end": {
                "expression": "after",
                "amount": delta_millis_end
            }
        }

        assert conditions is not None
        if len(conditions) == 0:
            raise ValueError('A trigger must contain at least one condition: you provided none')
        the_conditions = [dict(name=c.weather_param, expression=c.operator, amount=c.amount) for c in conditions]

        assert area is not None
        if len(area) == 0:
            raise ValueError('The area for a trigger must contain at least one geoJSON type: you provided none')
        the_area = [a.to_dict() for a in area]

        # >>> for the moment, no specific handling for alert channels

        status, payload = self.http_client.post(
            TRIGGERS_URI,
            params={'appid': self.API_key},
            data=dict(time_period=the_time_period, conditions=the_conditions, area=the_area),
            headers={'Content-Type': 'application/json'})
        return Trigger.from_dict(payload)

    def get_triggers(self):
        """
        Retrieves all of the user's triggers that are set on the Weather Alert API.

        :returns: list of `pyowm.alertapi30.trigger.Trigger` objects

        """
        status, data = self.http_client.get_json(
            TRIGGERS_URI,
            params={'appid': self.API_key},
            headers={'Content-Type': 'application/json'})
        return [Trigger.from_dict(item) for item in data]

    def get_trigger(self, trigger_id):
        """
        Retrieves the named trigger from the Weather Alert API.

        :param trigger_id: the ID of the trigger
        :type trigger_id: str
        :return: a `pyowm.alertapi30.trigger.Trigger` instance
        """
        assert isinstance(trigger_id, str), "Value must be a string"
        status, data = self.http_client.get_json(
            NAMED_TRIGGER_URI % trigger_id,
            params={'appid': self.API_key},
            headers={'Content-Type': 'application/json'})
        return Trigger.from_dict(data)

    def update_trigger(self, trigger):
        """
        Updates on the Alert API the trigger record having the ID of the specified Trigger object: the remote record is
        updated with data from the local Trigger object.

        :param trigger: the Trigger with updated data
        :type trigger: `pyowm.alertapi30.trigger.Trigger`
        :return: ``None`` if update is successful, an error otherwise
        """
        assert trigger is not None
        assert isinstance(trigger.id, str), "Value must be a string"
        the_time_period = {
            "start": {
                "expression": "after",
                "amount": trigger.start_after_millis
            },
            "end": {
                "expression": "after",
                "amount": trigger.end_after_millis
            }
        }
        the_conditions = [dict(name=c.weather_param, expression=c.operator, amount=c.amount) for c in trigger.conditions]
        the_area = [a.to_dict() for a in trigger.area]

        status, _ = self.http_client.put(
            NAMED_TRIGGER_URI % trigger.id,
            params={'appid': self.API_key},
            data=dict(time_period=the_time_period, conditions=the_conditions, area=the_area),
            headers={'Content-Type': 'application/json'})

    def delete_trigger(self, trigger):
        """
        Deletes from the Alert API the trigger record identified by the ID of the provided
        `pyowm.alertapi30.trigger.Trigger`, along with all related alerts

        :param trigger: the `pyowm.alertapi30.trigger.Trigger` object to be deleted
        :type trigger: `pyowm.alertapi30.trigger.Trigger`
        :returns: `None` if deletion is successful, an exception otherwise
        """
        assert trigger is not None
        assert isinstance(trigger.id, str), "Value must be a string"
        status, _ = self.http_client.delete(
            NAMED_TRIGGER_URI % trigger.id,
            params={'appid': self.API_key},
            headers={'Content-Type': 'application/json'})

    # ALERTS methods

    def get_alerts_for(self, trigger):
        """
        Retrieves all of the alerts that were fired for the specified Trigger
        :param trigger: the trigger
        :type trigger: `pyowm.alertapi30.trigger.Trigger`
        :return: list of `pyowm.alertapi30.alert.Alert` objects
        """
        assert trigger is not None
        assert isinstance(trigger.id, str), "Value must be a string"
        status, data = self.http_client.get_json(
            ALERTS_URI % trigger.id,
            params={'appid': self.API_key},
            headers={'Content-Type': 'application/json'})
        return [Alert.from_dict(item) for item in data]

    def get_alert(self, alert_id, trigger):
        """
        Retrieves info about the alert record on the Alert API that has the specified ID and belongs to the specified
        parent Trigger object
        :param trigger: the parent trigger
        :type trigger: `pyowm.alertapi30.trigger.Trigger`
        :param alert_id: the ID of the alert
        :type alert_id `pyowm.alertapi30.alert.Alert`
        :return: an `pyowm.alertapi30.alert.Alert` instance
        """
        assert trigger is not None
        assert alert_id is not None
        assert isinstance(alert_id, str), "Value must be a string"
        assert isinstance(trigger.id, str), "Value must be a string"
        status, data = self.http_client.get_json(
            NAMED_ALERT_URI % (trigger.id, alert_id),
            params={'appid': self.API_key},
            headers={'Content-Type': 'application/json'})
        return Alert.from_dict(data)

    def delete_all_alerts_for(self, trigger):
        """
        Deletes all of the alert that were fired for the specified Trigger
        :param trigger: the trigger whose alerts are to be cleared
        :type trigger: `pyowm.alertapi30.trigger.Trigger`
        :return: `None` if deletion is successful, an exception otherwise
        """
        assert trigger is not None
        assert isinstance(trigger.id, str), "Value must be a string"
        status, _ = self.http_client.delete(
            ALERTS_URI % trigger.id,
            params={'appid': self.API_key},
            headers={'Content-Type': 'application/json'})

    def delete_alert(self, alert):
        """
        Deletes the specified alert from the Alert API
        :param alert: the alert to be deleted
        :type alert: pyowm.alertapi30.alert.Alert`
        :return: ``None`` if the deletion was successful, an error otherwise
        """
        assert alert is not None
        assert isinstance(alert.id, str), "Value must be a string"
        assert isinstance(alert.trigger_id, str), "Value must be a string"
        status, _ = self.http_client.delete(
            NAMED_ALERT_URI % (alert.trigger_id, alert.id),
            params={'appid': self.API_key},
            headers={'Content-Type': 'application/json'})

    def __repr__(self):
        return '<%s.%s>' % (__name__, self.__class__.__name__)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.commons import exceptions


class Condition:
    """
    Object representing a condition to be checked on a specific weather parameter. A condition is given when comparing
    the weather parameter against a numerical value with respect to an operator.
    Allowed weather params and operators are specified by the `pyowm.utils.alertapi30.WeatherParametersEnum` and
    `pyowm.utils.alertapi30.OperatorsEnum` enumerator classes.
    :param weather_param: the weather variable to be checked (eg. TEMPERATURE, CLOUDS, ...)
    :type weather_param: str
    :param operator: the comparison operator to be applied to the weather variable (eg. GREATER_THAN, EQUAL, ...)
    :type operator: str
    :param amount: comparison value
    :type amount: int or float
    :param id: optional unique ID for this Condition instance
    :type id: str
    :returns:  a *Condition* instance
    :raises: *AssertionError* when either the weather param has wrong type or the operator has wrong type or the
    amount has wrong type

    """
    def __init__(self, weather_param, operator, amount, id=None):
        assert weather_param is not None
        assert isinstance(weather_param, str), "Value must be a string"
        self.weather_param = weather_param

        assert operator is not None
        assert isinstance(operator, str), "Value must be a string"
        self.operator = operator

        assert amount is not None
        assert isinstance(amount, (int, float))
        self.amount = amount
        self.id = id

    @classmethod
    def from_dict(cls, the_dict):
        if the_dict is None:
            raise exceptions.ParseAPIResponseError('Data is None')
        try:
            weather_param = the_dict['name']
            operator = the_dict['expression']
            amount = the_dict['amount']
            the_id = the_dict.get('_id', None)
            return Condition(weather_param, operator, amount, id=the_id)
        except KeyError as e:
            raise exceptions.ParseAPIResponseError('Impossible to parse data: %s' % e)

    def to_dict(self):
        return {
            'id': self.id,
            'weather_param': self.weather_param,
            'operator': self.operator,
            'amount': self.amount}

    def __repr__(self):
        return '<%s.%s - when %s %s %s>' % (__name__, self.__class__.__name__, self.weather_param,
                                            self.operator, str(self.amount))




#!/usr/bin/env python
# -*- coding: utf-8 -*-


ROOT_WEATHER_API = 'openweathermap.org/data/2.5'
OBSERVATION_URI = 'weather'
GROUP_OBSERVATIONS_URI = 'group'
STATION_URI = 'station'
FIND_OBSERVATIONS_URI = 'find'
FIND_STATION_URI = 'station/find'
BBOX_STATION_URI = 'box/station'
BBOX_CITY_URI = 'box/city'
THREE_HOURS_FORECAST_URI = 'forecast'
DAILY_FORECAST_URI = 'forecast/daily'
STATION_WEATHER_HISTORY_URI = 'history/station'
ONE_CALL_URI = 'onecall'
ONE_CALL_HISTORICAL_URI = ONE_CALL_URI + '/timemachine'
ONE_CALL_ROOT_URI = 'openweathermap.org/data/3.0'
ICONS_BASE_URI = 'http://openweathermap.org/img/wn/%s%s.png'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.commons import exceptions
from pyowm.utils import geo


class Location:
    """
    A class representing a location in the world. A location is defined through
    a toponym, a couple of geographic coordinates such as longitude and
    latitude and a numeric identifier assigned by the OWM Weather API that uniquely
    spots the location in the world. Optionally, the country specification may
    be provided.

    Further reference about OWM city IDs can be found at:
    http://bugs.openweathermap.org/projects/api/wiki/Api_2_5_weather#3-By-city-ID

    :param name: the location's toponym
    :type name: Unicode
    :param lon: the location's longitude, must be between -180.0 and 180.0
    :type lon: int/float
    :param lat: the location's latitude, must be between -90.0 and 90.0
    :type lat: int/float
    :param _id: the location's OWM city ID
    :type ID: int
    :param country: the location's country (``None`` by default)
    :type country: Unicode
    :returns: a *Location* instance
    :raises: *ValueError* if lon or lat values are provided out of bounds
    """

    def __init__(self, name, lon, lat, _id, country=None):
        self.name = name
        if lon is None or lat is None:
            raise ValueError("Either 'lon' or 'lat' must be specified")
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        self.lon = float(lon)
        self.lat = float(lat)
        self.id = _id
        self.country = country

    def to_geopoint(self):
        """
        Returns the geoJSON compliant representation of this location

        :returns: a ``pyowm.utils.geo.Point`` instance

        """
        if self.lon is None or self.lat is None:
            return None
        return geo.Point(self.lon, self.lat)

    @classmethod
    def from_dict(cls, the_dict):
        """
        Parses a *Location* instance out of a data dictionary. Only certain properties of the data dictionary
        are used: if these properties are not found or cannot be parsed, an exception is issued.

        :param the_dict: the input dictionary
        :type the_dict: `dict`
        :returns: a *Location* instance or ``None`` if no data is available
        :raises: *ParseAPIResponseError* if it is impossible to find or parse the data needed to build the result

        """
        if the_dict is None:
            raise exceptions.ParseAPIResponseError('Data is None')
        country = None
        if 'sys' in the_dict and 'country' in the_dict['sys']:
            country = the_dict['sys']['country']
        data = the_dict['city'] if 'city' in the_dict else the_dict
        name = data['name'] if 'name' in data else None
        ID = int(data['id']) if 'id' in data else None
        if 'coord' in data:
            lon = data['coord'].get('lon', 0.0)
            lat = data['coord'].get('lat', 0.0)
        elif 'station' in data and 'coord' in data['station']:
            if 'lon' in data['station']['coord']:
                lon = data['station']['coord'].get('lon', 0.0)
            elif 'lng' in data['station']['coord']:
                lon = data['station']['coord'].get('lng', 0.0)
            else:
                lon = 0.0
            lat = data['station']['coord'].get('lat', 0.0)
        elif 'lat' in the_dict and 'lon' in the_dict:
            lat = the_dict['lat']
            lon = the_dict['lon']
        else:
            raise KeyError("Impossible to read geographical coordinates from JSON")
        if 'country' in data:
            country = data['country']
        return Location(name, lon, lat, ID, country)

    def to_dict(self):
        """Dumps object to a dictionary

        :returns: a `dict`

        """
        return {'name': self.name,
                'coordinates': {'lon': self.lon, 'lat': self.lat},
                'ID': self.id,
                'country': self.country}

    def __repr__(self):
        return "<%s.%s - id=%s, name=%s, lon=%s, lat=%s>" % (__name__, \
                                                             self.__class__.__name__, self.id, self.name, str(self.lon), \
                                                             str(self.lat))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.utils import formatting
from pyowm.commons import exceptions


class NationalWeatherAlert:
    """
    A class representing a national weather alert . Alerts are provided by major national
    weather warning systems.

    This is the list of such major warning systems:
    https://openweathermap.org/api/one-call-api#listsource

    :param sender: the warning systems's name
    :type sender: str
    :param title: alert event name
    :type title: str
    :param description: description of the alert
    :type description: str
    :param start_time: UTC UNIXtime telling when the event is to start
    :type start_time: int
    :param end_time: UTC UNIXtime telling when the event is to end
    :type end_time: int
    :param tags: list of labels categorizing the event
    :type tags: list of str

    :returns: a *NationalWeatherAlert* instance
    :raises: *ValueError* if any parameter has wrong type
    """
    def __init__(self, sender, title, description, start_time, end_time, tags=None):
        assert title, 'Event title must be specified'
        assert start_time, 'Event start time must be specified'
        assert end_time, 'Event end time must be specified'

        self.sender = sender
        self.title = title
        self.description = description
        self.start = start_time
        self.end = end_time
        if tags is None:
            self.tags = []
        else:
            if not isinstance(tags, list):
                raise ValueError('If provided, event tags must be a list of strings')
            self.tags = tags

    def start_time(self, timeformat='unix'):
        """
        Returns the UTC start time of the weather alert event

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time
            '*iso*' for ISO8601-formatted string in the format ``YYYY-MM-DD HH:MM:SS+00``
            '*date* for ``datetime.datetime`` object instance
        :type timeformat: str
        :returns: an int or a str
        :raises: ValueError when negative values are provided

        """
        return formatting.timeformat(self.start, timeformat)

    def end_time(self, timeformat='unix'):
        """
        Returns the UTC end time of the weather alert event

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time
            '*iso*' for ISO8601-formatted string in the format ``YYYY-MM-DD HH:MM:SS+00``
            '*date* for ``datetime.datetime`` object instance
        :type timeformat: str
        :returns: an int or a str
        :raises: ValueError when negative values are provided

        """
        return formatting.timeformat(self.end, timeformat)

    @classmethod
    def from_dict(cls, the_dict):
        """
        Parses a *NationalWeatherAlert* instance out of a data dictionary. Only certain properties of the data dictionary
        are used: if these properties are not found or cannot be parsed, an exception is issued.

        :param the_dict: the input dictionary
        :type the_dict: `dict`
        :returns: a *NationalWeatherAlert* instance or ``None`` if no data is available
        :raises: *ParseAPIResponseError* if it is impossible to find or parse the data needed to build the result

        """
        if the_dict is None:
            raise exceptions.ParseAPIResponseError('Data is None')

        try:
            sender = the_dict['sender_name']
            title = the_dict['event']
            description = the_dict['description']
            start_time = the_dict['start']
            end_time = the_dict['end']
        except KeyError:
            raise exceptions.ParseAPIResponseError('Invalid data payload')
        tags = the_dict.get('tags', [])
        return NationalWeatherAlert(sender, title, description, start_time, end_time, tags)

    def to_dict(self):
        """Dumps object to a dictionary

        :returns: a `dict`
        """
        return {'sender_name': self.sender,
                'event': self.title,
                'start': self.start,
                'end': self.end,
                'description': self.description,
                'tags': self.tags }

    def __repr__(self):
        return "<%s.%s - sender=%s, title=%s, start=%s, end=%s>" % (
            __name__, self.__class__.__name__,
            self.sender, self.title, self.start_time('iso'), self.end_time('iso')
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
from pyowm.commons import exceptions
from pyowm.utils import formatting


class StationHistory:

    """
    A class representing historic weather measurements collected by a
    meteostation. Three types of historic meteostation data can be obtained by
    the OWM Weather API: ticks (one data chunk per minute) data, hourly and daily
    data.

    :param station_id: the numeric ID of the meteostation
    :type station_id: int
    :param interval: the time granularity of the meteostation data history
    :type interval: str
    :param reception_time: GMT UNIXtime of the data reception from the OWM web
         API
    :type reception_time: int
    :param measurements: a dictionary containing raw weather measurements
    :type measurements: dict
    :returns: a *StationHistory* instance
    :raises: *ValueError* when the supplied value for reception time is
        negative
    """

    def __init__(self, station_id, interval, reception_time, measurements):
        self.station_id = station_id
        self.interval = interval
        if reception_time < 0:
            raise ValueError("'reception_time' must be greater than 0")
        self.rec_time = reception_time
        self.measurements = measurements

    def reception_time(self, timeformat='unix'):
        """Returns the GMT time telling when the meteostation history data was
           received from the OWM Weather API

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time
            '*iso*' for ISO8601-formatted string in the format ``YYYY-MM-DD HH:MM:SS+00``
            '*date* for ``datetime.datetime`` object instance
        :type timeformat: str
        :returns: an int or a str
        :raises: ValueError

        """
        return formatting.timeformat(self.rec_time, timeformat)

    @classmethod
    def from_dict(cls, d):
        """
        Parses a *StationHistory* instance out of a data dictionary. Only certain properties of the data dictionary
        are used: if these properties are not found or cannot be parsed, an exception is issued.

        :param the_dict: the input dictionary
        :type the_dict: `dict`
        :returns: a *StationHistory* instance or ``None`` if no data is available
        :raises: *ParseAPIResponseError* if it is impossible to find or parse the
            data needed to build the result, *APIResponseError* if the input dict embeds an HTTP status error

        """
        if d is None:
            raise exceptions.ParseAPIResponseError('Data is None')
        # Check if server returned errors: this check overcomes the lack of use
        # of HTTP error status codes by the OWM API but it's supposed to be
        # deprecated as soon as the API implements a correct HTTP mechanism for
        # communicating errors to the clients. In addition, in this specific
        # case the OWM API responses are the very same either when no results
        # are found for a station and when the station does not exist!
        measurements = {}
        try:
            if 'cod' in d and d['cod'] != "200":
                raise exceptions.APIResponseError(
                                          "OWM API: error - response payload: " + str(d), d['cod'])
            if str(d['cnt']) == "0":
                return None
            else:
                for item in d['list']:
                    if 'temp' not in item:
                        temp = None
                    elif isinstance(item['temp'], dict):
                        temp = item['temp']['v']
                    else:
                        temp = item['temp']
                    if 'humidity' not in item:
                        hum = None
                    elif isinstance(item['humidity'], dict):
                        hum = item['humidity']['v']
                    else:
                        hum = item['humidity']
                    if 'pressure' not in item:
                        pres = None
                    elif isinstance(item['pressure'], dict):
                        pres = item['pressure']['v']
                    else:
                        pres = item['pressure']
                    if 'rain' in item and isinstance(item['rain']['today'],
                                                     dict):
                        rain = item['rain']['today']['v']
                    else:
                        rain = None
                    if 'wind' in item and isinstance(item['wind']['speed'],
                                                     dict):
                        wind = item['wind']['speed']['v']
                    else:
                        wind = None
                    measurements[item['dt']] = {"temperature": temp,
                                                "humidity": hum,
                                                "pressure": pres,
                                                "rain": rain,
                                                "wind": wind}
        except KeyError:
            raise exceptions.ParseAPIResponseError(__name__ + ': impossible to read input data')
        current_time = int(time.time())
        return StationHistory(None, None, current_time, measurements)

    def to_dict(self):
        """Dumps object to a dictionary

        :returns: a `dict`

        """
        return {"station_ID": self.station_id,
                "interval": self.interval,
                "reception_time": self.rec_time,
                "measurements": self.measurements}

    def __repr__(self):
        return "<%s.%s - station_id=%s, interval=%s>" % (__name__, self.__class__.__name__, self.station_id, self.interval)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from pyowm.commons import exceptions
from pyowm.commons.enums import ImageTypeEnum


class HttpRequestBuilder:

    URL_TEMPLATE_WITH_SUBDOMAINS = '{}://{}.{}/{}'
    URL_TEMPLATE_WITHOUT_SUBDOMAINS = '{}://{}/{}'

    """
    A stateful HTTP URL, params and headers builder with a fluent interface
    """
    def __init__(self, root_uri_token, api_key, config, has_subdomains=True):
        assert isinstance(root_uri_token, str)
        self.root = root_uri_token
        assert isinstance(api_key, str)
        self.api_key = api_key
        assert isinstance(config, dict)
        self.config = config
        assert isinstance(has_subdomains, bool)
        self.has_subdomains = has_subdomains
        self.schema = None
        self.subdomain = None
        self.proxies = None
        self.path = None
        self.params = {}
        self.headers = {}
        self._set_schema()
        self._set_subdomain()
        self._set_proxies()

    def _set_schema(self):
        use_ssl = self.config['connection']['use_ssl']
        self.schema = 'https' if use_ssl else 'http'

    def _set_subdomain(self):
        if self.has_subdomains:
            st = self.config['subscription_type']
            self.subdomain = st.subdomain

    def _set_proxies(self):
        if self.config['connection']['use_proxy']:
            self.proxies = self.config['proxies']
        else:
            self.proxies = {}

    def with_path(self, path_uri_token):
        assert isinstance(path_uri_token, str)
        self.path = path_uri_token
        return self

    def with_headers(self, headers):
        assert isinstance(headers, dict)
        self.headers.update(headers)
        return self

    def with_header(self, key, value):
        assert isinstance(key, str)
        try:
            json.dumps(value)
        except TypeError:
            raise ValueError('Header value is not JSON serializable')
        self.headers.update({key: value})
        return self

    def with_query_params(self, query_params):
        assert isinstance(query_params, dict)
        self.params.update(query_params)
        return self

    def with_api_key(self):
        self.params['APPID'] = self.api_key
        return self

    def with_language(self):
        self.params['lang'] = self.config['language']
        return self

    def build(self):
        if self.has_subdomains:
            return self.URL_TEMPLATE_WITH_SUBDOMAINS.format(self.schema, self.subdomain, self.root, self.path), \
                   self.params, self.headers, self.proxies
        else:
            return self.URL_TEMPLATE_WITHOUT_SUBDOMAINS.format(self.schema, self.root, self.path), \
                   self.params, self.headers, self.proxies

    def __repr__(self):
        return "<%s.%s>" % (__name__, self.__class__.__name__)


class HttpClient:

    """
    An HTTP client encapsulating some config data and abstarcting away data raw retrieval

    :param api_key: the OWM API key
    :type api_key: str
    :param config: the configuration dictionary (if not provided, a default one will be used)
    :type config: dict
    :param root_uri: the root URI of the API endpoint
    :type root_uri: str
    :param admits_subdomains: if the root URI of the API endpoint admits subdomains based on the subcription type (default: True)
    :type admits_subdomains: bool
    """

    def __init__(self, api_key, config, root_uri, admits_subdomains=True):
        assert isinstance(api_key, str)
        self.api_key = api_key
        assert isinstance(config, dict)
        self.config = config
        assert isinstance(root_uri, str)
        self.root_uri = root_uri
        assert isinstance(admits_subdomains, bool)
        self.admits_subdomains = admits_subdomains

        if self.config['connection']['max_retries'] is not None:
            # this adapter tells how to perform retries
            self.session_adapter = HTTPAdapter(
                max_retries=Retry(
                    total=self.config['connection']['max_retries'],
                    status_forcelist=[429, 500, 502, 503, 504],
                    allowed_methods =["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE"]
                )
            )
            # this is the adapted requests client
            self.http = requests.Session()
            self.http.mount("https://", self.session_adapter)
            self.http.mount("http://", self.session_adapter)
        else:
            self.http = requests

    def get_json(self, path, params=None, headers=None):
        builder = HttpRequestBuilder(self.root_uri, self.api_key, self.config, has_subdomains=self.admits_subdomains)\
            .with_path(path)\
            .with_api_key()\
            .with_language()\
            .with_query_params(params if params is not None else dict())\
            .with_headers(headers if headers is not None else dict())
        url, params, headers, proxies = builder.build()
        try:
            resp = self.http.get(url, params=params, headers=headers, proxies=proxies,
                                timeout=self.config['connection']['timeout_secs'],
                                verify=self.config['connection']['verify_ssl_certs'])
        except requests.exceptions.SSLError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.ConnectionError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.Timeout:
            raise exceptions.TimeoutError('API call timed out')
        HttpClient.check_status_code(resp.status_code, resp.text)
        try:
            return resp.status_code, resp.json()
        except:
            raise exceptions.ParseAPIResponseError('Impossible to parse API response data')

    def get_png(self, path, params=None, headers=None):
        # check URL fromt the metaimage: if it looks like a complete URL, use that one (I know, it's a hack...)
        try:
            partial_path = path.split(self.root_uri)[1].lstrip('/')
        except:
            partial_path = path

        builder = HttpRequestBuilder(self.root_uri, self.api_key, self.config, has_subdomains=self.admits_subdomains)\
            .with_path(partial_path)\
            .with_api_key()\
            .with_language()\
            .with_query_params(params if params is not None else dict())\
            .with_headers(headers if headers is not None else dict())\
            .with_header('Accept', ImageTypeEnum.PNG.mime_type)
        url, params, headers, proxies = builder.build()
        try:
            resp = self.http.get(url, stream=True, params=params, headers=headers, proxies=proxies,
                                timeout=self.config['connection']['timeout_secs'],
                                verify=self.config['connection']['verify_ssl_certs'])
        except requests.exceptions.SSLError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.ConnectionError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.Timeout:
            raise exceptions.TimeoutError('API call timeouted')
        HttpClient.check_status_code(resp.status_code, resp.text)
        try:
            return resp.status_code, resp.content
        except:
            raise exceptions.ParseAPIResponseError('Impossible to parse'
                                                          'API response data')

    def get_geotiff(self, path, params=None, headers=None):
        # check URL fromt the metaimage: if it looks like a complete URL, use that one (I know, it's a hack...)
        try:
            partial_path = path.split(self.root_uri)[1].lstrip('/')
        except:
            partial_path = path

        builder = HttpRequestBuilder(self.root_uri, self.api_key, self.config, has_subdomains=self.admits_subdomains)\
            .with_path(partial_path)\
            .with_api_key()\
            .with_language()\
            .with_query_params(params if params is not None else dict())\
            .with_headers(headers if headers is not None else dict())\
            .with_header('Accept', ImageTypeEnum.GEOTIFF.mime_type)
        url, params, headers, proxies = builder.build()
        try:
            resp = self.http.get(url, stream=True, params=params, headers=headers, proxies=proxies,
                                timeout=self.config['connection']['timeout_secs'],
                                verify=self.config['connection']['verify_ssl_certs'])
        except requests.exceptions.SSLError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.ConnectionError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.Timeout:
            raise exceptions.TimeoutError('API call timeouted')
        HttpClient.check_status_code(resp.status_code, resp.text)
        try:
            return resp.status_code, resp.content
        except:
            raise exceptions.ParseAPIResponseError('Impossible to parse'
                                                          'API response data')

    def post(self, path, params=None, data=None, headers=None):
        builder = HttpRequestBuilder(self.root_uri, self.api_key, self.config, has_subdomains=self.admits_subdomains)\
            .with_path(path)\
            .with_api_key()\
            .with_language()\
            .with_query_params(params if params is not None else dict())\
            .with_headers(headers if headers is not None else dict())
        url, params, headers, proxies = builder.build()
        try:
            resp = self.http.post(url, params=params, json=data, headers=headers, proxies=proxies,
                                 timeout=self.config['connection']['timeout_secs'],
                                 verify=self.config['connection']['verify_ssl_certs'])
        except requests.exceptions.SSLError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.ConnectionError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.Timeout:
            raise exceptions.TimeoutError('API call timeouted')
        HttpClient.check_status_code(resp.status_code, resp.text)
        # this is a defense against OWM API responses containing an empty body!
        try:
            json_data = resp.json()
        except:
            json_data = {}
        return resp.status_code, json_data

    def put(self, path, params=None, data=None, headers=None):
        builder = HttpRequestBuilder(self.root_uri, self.api_key, self.config, has_subdomains=self.admits_subdomains)\
            .with_path(path)\
            .with_api_key()\
            .with_language()\
            .with_query_params(params if params is not None else dict())\
            .with_headers(headers if headers is not None else dict())
        url, params, headers, proxies = builder.build()
        try:
            resp = self.http.put(url, params=params, json=data, headers=headers, proxies=proxies,
                                timeout=self.config['connection']['timeout_secs'],
                                verify=self.config['connection']['verify_ssl_certs'])
        except requests.exceptions.SSLError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.ConnectionError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.Timeout:
            raise exceptions.TimeoutError('API call timeouted')
        HttpClient.check_status_code(resp.status_code, resp.text)
        # this is a defense against OWM API responses containing an empty body!
        try:
            json_data = resp.json()
        except:
            json_data = {}
        return resp.status_code, json_data

    def delete(self, path, params=None, data=None, headers=None):
        builder = HttpRequestBuilder(self.root_uri, self.api_key, self.config, has_subdomains=self.admits_subdomains)\
            .with_path(path)\
            .with_api_key()\
            .with_language()\
            .with_query_params(params if params is not None else dict())\
            .with_headers(headers if headers is not None else dict())
        url, params, headers, proxies = builder.build()
        try:
            resp = self.http.delete(url, params=params, json=data, headers=headers, proxies=proxies,
                                   timeout=self.config['connection']['timeout_secs'],
                                   verify=self.config['connection']['verify_ssl_certs'])
        except requests.exceptions.SSLError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.ConnectionError as e:
            raise exceptions.InvalidSSLCertificateError(str(e))
        except requests.exceptions.Timeout:
            raise exceptions.TimeoutError('API call timeouted')
        HttpClient.check_status_code(resp.status_code, resp.text)
        # this is a defense against OWM API responses containing an empty body!
        try:
            json_data = resp.json()
        except:
            json_data = None
        return resp.status_code, json_data

    @classmethod
    def check_status_code(cls, status_code, payload):
        if status_code < 400:
            return
        if status_code == 400 or status_code not in [401, 404, 502]:
            raise exceptions.APIRequestError(payload)
        elif status_code == 401:
            raise exceptions.UnauthorizedError('Invalid API Key provided')
        elif status_code == 404:
            raise exceptions.NotFoundError('Unable to find the resource')
        else:
            raise exceptions.BadGatewayError('Unable to contact the upstream server')

    def __repr__(self):
        return "<%s.%s - root: %s>" % (__name__, self.__class__.__name__, self.root_uri)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.commons.enums import ImageTypeEnum
from pyowm.commons.image import Image
from pyowm.commons.tile import Tile
from pyowm.utils import formatting


class MetaImage:
    """
    A class representing metadata for a satellite-acquired image

    :param url: the public URL of the image
    :type url: str
    :param preset: the preset of the image (supported values are listed by `pyowm.agroapi10.enums.PresetEnum`)
    :type preset: str
    :param satellite_name: the name of the satellite that acquired the image (supported values are listed
        by `pyowm.agroapi10.enums.SatelliteEnum`)
    :type satellite_name: str
    :param acquisition_time: the UTC Unix epoch when the image was acquired
    :type acquisition_time: int
    :param valid_data_percentage: approximate percentage of valid data coverage
    :type valid_data_percentage: float
    :param cloud_coverage_percentage: approximate percentage of cloud coverage on the scene
    :type cloud_coverage_percentage: float
    :param sun_azimuth: sun azimuth angle at scene acquisition time
    :type sun_azimuth: float
    :param sun_elevation: sun zenith angle at scene acquisition time
    :type sun_elevation: float
    :param polygon_id: optional id of the polygon the image refers to
    :type polygon_id: str
    :param stats_url: the public URL of the image statistics, if available
    :type stats_url: str or `None`
    :returns: an `MetaImage` object
    """

    image_type = None

    def __init__(self, url, preset, satellite_name, acquisition_time,
                 valid_data_percentage, cloud_coverage_percentage, sun_azimuth, sun_elevation, polygon_id=None,
                 stats_url=None):
        assert isinstance(url, str)
        self.url = url
        self.preset = preset
        self.satellite_name = satellite_name
        assert isinstance(acquisition_time, int)
        assert acquisition_time >= 0, 'acquisition_time cannot be negative'
        self._acquisition_time = acquisition_time
        assert isinstance(valid_data_percentage, (float, int))
        assert valid_data_percentage >= 0., 'valid_data_percentage cannot be negative'
        self.valid_data_percentage = valid_data_percentage
        assert isinstance(cloud_coverage_percentage, (float, int))
        assert cloud_coverage_percentage >= 0., 'cloud_coverage_percentage cannot be negative'
        self.cloud_coverage_percentage = cloud_coverage_percentage
        assert isinstance(sun_azimuth, (float, int))
        assert sun_azimuth >= 0. and sun_azimuth <= 360., 'sun_azimuth must be between 0 and 360 degrees'
        self.sun_azimuth  = sun_azimuth
        assert isinstance(sun_elevation, (float, int))
        assert sun_elevation >= 0. and sun_elevation <= 90., 'sun_elevation must be between 0 and 90 degrees'
        self.sun_elevation = sun_elevation
        self.polygon_id = polygon_id
        self.stats_url = stats_url

    def acquisition_time(self, timeformat='unix'):
        """Returns the UTC time telling when the image data was acquired by the satellite

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time
            '*iso*' for ISO8601-formatted string in the format ``YYYY-MM-DD HH:MM:SS+00``
            '*date* for ``datetime.datetime`` object instance
        :type timeformat: str
        :returns: an int or a str

        """
        return formatting.timeformat(self._acquisition_time, timeformat)

    def __repr__(self):
        return "<%s.%s - %s %s image acquired at %s by %s on polygon with id=%s>" % (
            __name__, self.__class__.__name__,
            self.image_type.name if self.image_type is not None else '',
            self.preset, self.acquisition_time('iso'), self.satellite_name,
            self.polygon_id if self.polygon_id is not None else 'None')


class MetaPNGImage(MetaImage):
    """
    Class representing metadata for a satellite image of a polygon in PNG format
    """
    image_type = ImageTypeEnum.PNG


class MetaTile(MetaImage):
    """
    Class representing metadata for a tile in PNG format
    """
    image_type = ImageTypeEnum.PNG


class MetaGeoTiffImage(MetaImage):
    """
    Class representing metadata for a satellite image of a polygon in GeoTiff format
    """
    image_type = ImageTypeEnum.GEOTIFF


class SatelliteImage:
    """
    Class representing a downloaded satellite image, featuring both metadata and data

    :param metadata: the metadata for this satellite image
    :type metadata: a `pyowm.agro10.imagery.MetaImage` subtype instance
    :param data: the actual data for this satellite image
    :type data: either `pyowm.commons.image.Image` or `pyowm.commons.tile.Tile` object
    :param downloaded_on: the UNIX epoch this satellite image was downloaded at
    :type downloaded_on: int or `None`
    :param palette: ID of the color palette of the downloaded images. Values are provided by `pyowm.agroapi10.enums.PaletteEnum`
    :type palette: str or `None`
    :returns: a `pyowm.agroapi10.imagery.SatelliteImage` instance
    """

    def __init__(self, metadata, data, downloaded_on=None, palette=None):
        assert isinstance(metadata, MetaImage)
        self.metadata = metadata
        assert isinstance(data, (Image, Tile))
        self.data = data
        if downloaded_on is not None:
            assert isinstance(downloaded_on, int)
            self._downloaded_on = downloaded_on
        if palette is not None:
            assert isinstance(palette, str)
            self.palette = palette

    def downloaded_on(self, timeformat='unix'):
        """Returns the UTC time telling when the satellite image was downloaded from the OWM Agro API

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time
            '*iso*' for ISO8601-formatted string in the format ``YYYY-MM-DD HH:MM:SS+00``
            '*date* for ``datetime.datetime`` object instance
        :type timeformat: str
        :returns: an int or a str

        """
        return formatting.timeformat(self._downloaded_on, timeformat)

    def persist(self, path_to_file):
        """
        Saves the satellite image to disk on a file

        :param path_to_file: path to the target file
        :type path_to_file: str
        :return: `None`
        """
        self.data.persist(path_to_file)

    def __repr__(self):
        return "<%s.%s - %s %s satellite image downloaded on: %s>" % (
            __name__, self.__class__.__name__,
            self.metadata.preset, self.metadata.satellite_name,
            self.downloaded_on('iso') if self._downloaded_on is not None else 'None')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import json

from pyowm.commons import exceptions
from pyowm.config import DEFAULT_CONFIG
from pyowm.commons.enums import SubscriptionTypeEnum


def get_config_from(path_to_file):
    """
    Loads configuration data from the supplied file and returns it.

    :param path_to_file: path to the configuration file
    :type path_to_file: str
    :returns: the configuration `dict`
    :raises: `ConfigurationNotFoundError` when the supplied filepath is not a regular file; `ConfigurationParseError`
        when the supplied file cannot be parsed
    """
    assert path_to_file is not None
    if not os.path.isfile(path_to_file):
        raise exceptions.ConfigurationNotFoundError(
            'Configuration file not found: {}'.format(path_to_file))
    with open(path_to_file, 'r') as cf:
        try:
            config_data = json.load(cf)
            config_data['subscription_type'] = SubscriptionTypeEnum.lookup_by_name(config_data['subscription_type'])
            return config_data
        except Exception:
            raise exceptions.ConfigurationParseError()


def get_default_config():
    """
    Returns the default PyOWM configuration.

    :returns: the configuration `dict`
    """
    return DEFAULT_CONFIG


def get_default_config_for_subscription_type(name):
    """
    Returns the PyOWM configuration for a specific OWM API Plan subscription type

    :param name: name of the subscription type
    :type name: str
    :returns: the configuration `dict`
    """
    assert isinstance(name, str)
    config = get_default_config()
    config['subscription_type'] = SubscriptionTypeEnum.lookup_by_name(name)
    return config


def get_default_config_for_proxy(http_url, https_url):
    """
    Returns the PyOWM configuration to be used behind a proxy server

    :param http_url: URL connection string for HTTP protocol
    :type http_url: str
    :param https_url: URL connection string for HTTPS protocol
    :type https_url: str
    :returns: the configuration `dict`
    """
    assert isinstance(http_url, str)
    assert isinstance(https_url, str)
    config = get_default_config()
    config['connection']['use_proxy'] = True
    config['proxies']['http'] = http_url
    config['proxies']['https'] = https_url
    return config



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.alertapi30.condition import Condition
from pyowm.commons import exceptions
from pyowm.utils import formatting


class AlertChannel:
    """
    Base class representing a channel through which one can acknowledge that a weather alert has been issued.
    Examples: OWM API polling, push notifications, email notifications, etc.
    This feature is yet to be implemented by the OWM API.
    :param name: name of the channel
    :type name: str
    :returns: an *AlertChannel* instance

    """
    def __init__(self, name):
        self.name = name

    def to_dict(self):
        return dict(name=self.name)

    def __repr__(self):
        return '<%s.%s - name: %s>' % (__name__, self.__class__.__name__, self.name)


class Alert:
    """
    Represents the situation happening when any of the conditions bound to a `Trigger` is met. Whenever this happens, an
    `Alert` object is created (or updated) and is bound to its parent `Trigger`. The trigger can then be polled to check
    what alerts have been fired on it.
    :param id: unique alert identifier
    :type name: str
    :param trigger_id: link back to parent `Trigger`
    :type trigger_id: str
    :param met_conditions: list of dict, each one referring to a `Condition` obj bound to the parent `Trigger` and reporting
    the actual measured values that made this `Alert` fire
    :type met_conditions: list of dict
    :param coordinates: dict representing the geocoordinates where the `Condition` triggering the `Alert` was met
    :type coordinates: dict
    :param last_update: epoch of the last time when this `Alert` has been fired
    :type last_update: int

    """
    def __init__(self, id, trigger_id, met_conditions, coordinates, last_update=None):
        assert id is not None
        assert isinstance(id, str), "Value must be a string"
        self.id = id

        assert trigger_id is not None
        assert isinstance(trigger_id, str), "Value must be a string"
        self.trigger_id = trigger_id

        assert met_conditions is not None
        assert isinstance(met_conditions, list)
        self.met_conditions = met_conditions

        assert coordinates is not None
        assert isinstance(coordinates, dict)
        self.coordinates = coordinates

        if last_update is not None:
            assert isinstance(last_update, int)
        self.last_update = last_update

    @classmethod
    def from_dict(cls, the_dict):
        """
        Parses a *Alert* instance out of a data dictionary. Only certain properties of the data dictionary
        are used: if these properties are not found or cannot be parsed, an exception is issued.

        :param the_dict: the input dictionary
        :type the_dict: `dict`
        :returns: a *Alert* instance or ``None`` if no data is available
        :raises: *ParseAPIResponseError* if it is impossible to find or parse the data needed to build the result

        """
        if the_dict is None:
            raise exceptions.ParseAPIResponseError('Data is None')
        try:
            alert_id = the_dict['_id']
            t = the_dict['last_update'].split('.')[0].replace('T', ' ') + '+00:00'
            alert_last_update = formatting.ISO8601_to_UNIXtime(t)
            alert_trigger_id = the_dict['triggerId']
            alert_met_conds = [
                dict(current_value=c['current_value']['min'], condition=Condition.from_dict(c['condition']))
                    for c in the_dict['conditions']
            ]
            alert_coords = the_dict['coordinates']
            return Alert(alert_id, alert_trigger_id, alert_met_conds, alert_coords, last_update=alert_last_update)
        except ValueError as e:
            raise exceptions.ParseAPIResponseError('Impossible to parse JSON: %s' % e)
        except KeyError as e:
            raise exceptions.ParseAPIResponseError('Impossible to parse JSON: %s' % e)

    def to_dict(self):
        """Dumps object to a dictionary

        :returns: a `dict`

        """
        return {
            'id': self.id,
            'trigger_id': self.trigger_id,
            'met_conditions': self.met_conditions,
            'coordinates': self.coordinates,
            'last_update': self.last_update}

    def __repr__(self):
        return "<%s.%s - id=%s, trigger id=%s, last update=%s>" % (
            __name__,
            self.__class__.__name__,
            self.id,
            self.trigger_id,
            formatting.to_ISO8601(self.last_update))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.alertapi30.alert import AlertChannel


class WeatherParametersEnum:
    """
    Allowed weather parameters for condition checking

    """
    TEMPERATURE = 'temp'  # Kelvin
    PRESSURE = 'pressure'
    HUMIDITY = 'humidity'
    WIND_SPEED = 'wind_speed'
    WIND_DIRECTION = 'wind_direction'
    CLOUDS = 'clouds'  # Coverage %

    @classmethod
    def items(cls):
        """
        All values for this enum
        :return: list of str

        """
        return [
            cls.TEMPERATURE,
            cls.PRESSURE,
            cls.HUMIDITY,
            cls.WIND_SPEED,
            cls.WIND_DIRECTION,
            cls.CLOUDS
        ]


class OperatorsEnum:
    """
    Allowed comparison operators for condition checking upon weather parameters

    """
    GREATER_THAN = '$gt'
    GREATER_THAN_EQUAL = '$gte'
    LESS_THAN = '$lt'
    LESS_THAN_EQUAL = '$lte'
    EQUAL = '$eq'
    NOT_EQUAL = '$ne'

    @classmethod
    def items(cls):
        """
        All values for this enum
        :return: list of str

        """
        return [
            cls.GREATER_THAN,
            cls.GREATER_THAN_EQUAL,
            cls.LESS_THAN,
            cls.LESS_THAN_EQUAL,
            cls.EQUAL,
            cls.NOT_EQUAL
        ]


class AlertChannelsEnum:
    """
    Allowed alert channels

    """
    OWM_API_POLLING = AlertChannel('OWM API POLLING')

    @classmethod
    def items(cls):
        """
        All values for this enum
        :return: list of str

        """
        return [
            cls.OWM_API_POLLING
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

ROOT_STATIONS_API_URL = 'openweathermap.org/data/3.0'

# Stations
STATIONS_URI = 'stations'
NAMED_STATION_URI = 'stations/%s'

# Measurements
MEASUREMENTS_URI = 'measurements'



#!/usr/bin/env python
# -*- coding: utf-8 -*-"""

# Temperature conversion constants
KELVIN_OFFSET = 273.15
FAHRENHEIT_OFFSET = 32.0
FAHRENHEIT_DEGREE_SCALE = 1.8

# Wind speed conversion constants
MILES_PER_HOUR_FOR_ONE_METER_PER_SEC = 2.23694
KM_PER_HOUR_FOR_ONE_METER_PER_SEC = 3.6
KNOTS_FOR_ONE_METER_PER_SEC = 1.94384

# Barometric conversion constants
HPA_FOR_ONE_INHG = 33.8639

# Visibility distance conversion constants
MILE_FOR_ONE_METER = 0.000621371
KMS_FOR_ONE_METER = .001

# Decimal precision
ROUNDED_TO = 2


def kelvin_dict_to(d, target_temperature_unit):
    """
    Converts all the values in a dict from Kelvin temperatures to the
    specified temperature format.

    :param d: the dictionary containing Kelvin temperature values
    :type d: dict
    :param target_temperature_unit: the target temperature unit, may be:
        'celsius' or 'fahrenheit'
    :type target_temperature_unit: str
    :returns: a dict with the same keys as the input dict and converted
        temperature values as values
    :raises: *ValueError* when unknown target temperature units are provided

    """
    if target_temperature_unit == 'kelvin':
        return d
    elif target_temperature_unit == 'celsius':
        return {key: kelvin_to_celsius(d[key]) for key in d}
    elif target_temperature_unit == 'fahrenheit':
        return {key: kelvin_to_fahrenheit(d[key]) for key in d}
    else:
        raise ValueError("Invalid value for target temperature conversion \
                         unit")


def kelvin_to_celsius(kelvintemp):
    """
    Converts a numeric temperature from Kelvin degrees to Celsius degrees

    :param kelvintemp: the Kelvin temperature
    :type kelvintemp: int/long/float
    :returns: the float Celsius temperature
    :raises: *TypeError* when bad argument types are provided

    """
    if kelvintemp < 0:
        raise ValueError(__name__ +
                         ": negative temperature values not allowed")
    celsiustemp = kelvintemp - KELVIN_OFFSET
    return float("{0:.2f}".format(celsiustemp))


def kelvin_to_fahrenheit(kelvintemp):
    """
    Converts a numeric temperature from Kelvin degrees to Fahrenheit degrees

    :param kelvintemp: the Kelvin temperature
    :type kelvintemp: int/long/float
    :returns: the float Fahrenheit temperature

    :raises: *TypeError* when bad argument types are provided
    """
    if kelvintemp < 0:
        raise ValueError(__name__ +
                         ": negative temperature values not allowed")
    fahrenheittemp = (kelvintemp - KELVIN_OFFSET) * \
        FAHRENHEIT_DEGREE_SCALE + FAHRENHEIT_OFFSET
    return float("{0:.2f}".format(fahrenheittemp))


def metric_wind_dict_to_imperial(d):
    """
    Converts all the wind values in a dict from meters/sec (metric measurement
    system) to miles/hour (imperial measurement system)
    .

    :param d: the dictionary containing metric values
    :type d: dict
    :returns: a dict with the same keys as the input dict and values converted
        to miles/hour

    """
    result = {}
    for key, value in d.items():
        if key != 'deg':  # do not convert wind degree
            result[key] = value * MILES_PER_HOUR_FOR_ONE_METER_PER_SEC
        else:
            result[key] = value
    return result


def metric_wind_dict_to_km_h(d):
    """
    Converts all the wind values in a dict from meters/sec
    to km/hour.

    :param d: the dictionary containing metric values
    :type d: dict
    :returns: a dict with the same keys as the input dict and values converted
        to km/hour

    """
    result = {}
    for key, value in d.items():
        if key != 'deg':  # do not convert wind degree
            result[key] = value * KM_PER_HOUR_FOR_ONE_METER_PER_SEC
        else:
            result[key] = value
    return result


def metric_wind_dict_to_knots(d):
    """
    Converts all the wind values in a dict from meters/sec
    to knots

    :param d: the dictionary containing metric values
    :type d: dict
    :returns: a dict with the same keys as the input dict and values converted
        to km/hour

    """
    result = {}
    for key, value in d.items():
        if key != 'deg':  # do not convert wind degree
            result[key] = value * KNOTS_FOR_ONE_METER_PER_SEC
        else:
            result[key] = value
    return result


def metric_wind_dict_to_beaufort(d):
    """
    Converts all the wind values in a dict from meters/sec
    to the corresponding Beaufort scale level (which is not an exact number but rather
    represents a range of wind speeds - see: https://en.wikipedia.org/wiki/Beaufort_scale).
    Conversion table: https://www.windfinder.com/wind/windspeed.htm

    :param d: the dictionary containing metric values
    :type d: dict
    :returns: a dict with the same keys as the input dict and values converted
        to Beaufort level

    """
    result = {}
    for key, value in d.items():
        if key != 'deg':  # do not convert wind degree
            if value <= 0.2:
                bf = 0
            elif 0.2 < value <= 1.5:
                bf = 1
            elif 1.5 < value <= 3.3:
                bf = 2
            elif 3.3 < value <= 5.4:
                bf = 3
            elif 5.4 < value <= 7.9:
                bf = 4
            elif 7.9 < value <= 10.7:
                bf = 5
            elif 10.7 < value <= 13.8:
                bf = 6
            elif 13.8 < value <= 17.1:
                bf = 7
            elif 17.1 < value <= 20.7:
                bf = 8
            elif 20.7 < value <= 24.4:
                bf = 9
            elif 24.4 < value <= 28.4:
                bf = 10
            elif 28.4 < value <= 32.6:
                bf = 11
            else:
                bf = 12
            result[key] = bf
        else:
            result[key] = value
    return result


def metric_pressure_dict_to_inhg(d):
    """
    Converts all barometric pressure values in a dict to "inches of mercury."

    :param d: the dictionary containing metric values
    :type d: dict
    :returns: a dict with the same keys as the input dict and values converted
        to "Hg or inHg (inches of mercury)

    Note what OWM says about pressure: "Atmospheric pressure [is given in hPa]
    (on the sea level, if there is no sea_level or grnd_level data)"
    """
    result = dict()
    for key, value in d.items():
        if value is None:
            continue
        result[key] = round((value / HPA_FOR_ONE_INHG), ROUNDED_TO)
    return result


def visibility_distance_to(v, target_visibility_unit='kilometers'):
    """
    Converts visibility distance (in meters) to kilometers or miles
    Defaults to kilometer conversion

    :param distance: the value of visibility_distance
    :type distance: int
    :param target_visibility_unit: the unit of conversion
    :type target_visibility_unit: str
    :returns: a converted value for visibility_distance (float)
    """
    if v is None:
        return v

    if target_visibility_unit == 'kilometers':
        const = KMS_FOR_ONE_METER
    elif target_visibility_unit == 'miles':
        const = MILE_FOR_ONE_METER
    else:
        raise ValueError('Invalid value for target visibility distance unit')

    return round(v * const, ROUNDED_TO)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.commons import exceptions
from pyowm.utils import formatting, timestamps
from pyowm.weatherapi30 import location


class AirStatus:
    """
    A class representing a dataset about air quality

    :param reference_time: GMT UNIXtime telling when the data has been measured
    :type reference_time: int
    :param location: the *Location* relative to this measurement
    :type location: *Location*
    :param interval: the time granularity of the CO observation
    :type interval: str
    :param air_quality_data: the dataset
    :type air_quality_data: dict
    :param reception_time: GMT UNIXtime telling when the CO observation has
        been received from the OWM Weather API
    :type reception_time: int
    :returns: an *COIndex* instance
    :raises: *ValueError* when negative values are provided as reception time,
      CO samples are not provided in a list

    """

    def __init__(self, reference_time, location, air_quality_data, reception_time):
        if reference_time < 0:
            raise ValueError("'reference_time' must be greater than 0")
        self.ref_time = reference_time
        self.location = location
        if not isinstance(air_quality_data, dict):
            raise ValueError("'air_quality_data' must be a list")
        self.air_quality_data = air_quality_data
        for key, val in air_quality_data.items():
            setattr(self, key, val)
        if reception_time < 0:
            raise ValueError("'reception_time' must be greater than 0")
        self.rec_time = reception_time

    def reference_time(self, timeformat='unix'):
        """
        Returns the GMT time telling when the air quality data have been measured

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time
            '*iso*' for ISO8601-formatted string in the format ``YYYY-MM-DD HH:MM:SS+00:00``
            '*date* for ``datetime.datetime`` object instance
        :type timeformat: str
        :returns: an int or a str
        :raises: ValueError when negative values are provided

        """
        return formatting.timeformat(self.ref_time, timeformat)

    def reception_time(self, timeformat='unix'):
        """
        Returns the GMT time telling when the air quality data has been received
        from the OWM Weather API

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time
            '*iso*' for ISO8601-formatted string in the format ``YYYY-MM-DD HH:MM:SS+00:00``
            '*date* for ``datetime.datetime`` object instance
        :type timeformat: str
        :returns: an int or a str
        :raises: ValueError when negative values are provided

        """
        return formatting.timeformat(self.rec_time, timeformat)


    @classmethod
    def from_dict(cls, the_dict):
        """
        Parses an *AirStatus* instance or `list` of instances out of a data dictionary.

        :param the_dict: the input dictionary
        :type the_dict: `dict`
        :returns: a *AirStatus* instance or ``list` of such instances
        :raises: *ParseAPIResponseError* if it is impossible to find or parse the data needed to build the result

        """
        if the_dict is None:
            raise exceptions.ParseAPIResponseError('Data is None')
        try:
            # -- location
            lon = float(the_dict['coord']['lat'])
            lat = float(the_dict['coord']['lon'])
            place = location.Location(None, lon, lat, None)

            # -- reception time (now)
            rcp_time = timestamps.now('unix')

            def build_air_status(item_dict, location, reception_time):
                # -- reference time (strip away Z and T on ISO8601 format)
                reference_time = item_dict['dt']

                # -- air quality data
                data = item_dict['components']
                data['aqi'] = item_dict['main']['aqi']

                return AirStatus(reference_time, location, data, reception_time)

            items = the_dict['list']

            # one datapoint
            if len(items) == 1:
                return build_air_status(items[0], place, rcp_time)
            # multiple datapoints
            else:
                return [build_air_status(item, place, rcp_time) for item in items]

        except KeyError:
            raise exceptions.ParseAPIResponseError(
                      ''.join([__name__, ': impossible to parse AirStatus']))

    def to_dict(self):
        """Dumps object to a dictionary

        :returns: a `dict`

        """
        return {"reference_time": self.ref_time,
                "location": self.location.to_dict(),
                "air_quality_data": self.air_quality_data,
                "reception_time": self.rec_time}

    def __repr__(self):
        return "<%s.%s - reference time=%s, reception time=%s, location=%s" % (
                    __name__,
                    self.__class__.__name__,
                    self.reference_time('iso'),
                    self.reception_time('iso'),
                    str(self.location))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime, timedelta, timezone
from pyowm.utils import formatting


def now(timeformat='date'):
    """
    Returns the current time in the specified timeformat.

    :param timeformat: the target format for the time conversion. May be:
        '*date*' (default - outputs a ``datetime.datetime`` object), '*unix*'
        (outputs a long UNIXtime) or '*iso*' (outputs an ISO8601-formatted
        string with pattern ``YYYY-MM-DD HH:MM:SS+00``)
    :type timeformat: str
    :returns: the current time value
    :raises: ValueError when unknown timeformat switches are provided or
        when negative time values are provided
    """
    return formatting.timeformat(datetime.now(timezone.utc), timeformat)


def next_hour(date=None):
    """
    Gives the ``datetime.datetime`` object corresponding to the next hour
    from now or from the specified ``datetime.datetime`` object.

    :param date: the date you want an hour to be added (if left ``None``,
        the current date and time will be used)
    :type date: ``datetime.datetime`` object
    :returns: a ``datetime.datetime`` object
    """
    return _timedelta_hours(1, date)


def last_hour(date=None):
    """
    Gives the ``datetime.datetime`` object corresponding to the last hour
    before now or before the specified ``datetime.datetime`` object.

    :param date: the date you want an hour to be subtracted from (if left
        ``None``, the current date and time will be used)
    :type date: ``datetime.datetime`` object
    :returns: a ``datetime.datetime`` object
    """
    return _timedelta_hours(-1, date)


def next_three_hours(date=None):
    """
    Gives the ``datetime.datetime`` object corresponding to the next three
    hours from now or from the specified ``datetime.datetime`` object.

    :param date: the date you want three hours to be added (if left ``None``,
        the current date and time will be used)
    :type date: ``datetime.datetime`` object
    :returns: a ``datetime.datetime`` object
    """
    return _timedelta_hours(3, date)


def last_three_hours(date=None):
    """
    Gives the ``datetime.datetime`` object corresponding to last three
    hours before now or before the specified ``datetime.datetime`` object.

    :param date: the date you want three hours to be subtracted from (if left
        ``None``, the current date and time will be used)
    :type date: ``datetime.datetime`` object
    :returns: a ``datetime.datetime`` object
    """
    return _timedelta_hours(-3, date)


def tomorrow(hour=None, minute=None):
    """
    Gives the ``datetime.datetime`` object corresponding to tomorrow. The
    default value for optional parameters is the current value of hour and
    minute. I.e: when called without specifying values for parameters, the
    resulting object will refer to the time = now + 24 hours; when called with
    only hour specified, the resulting object will refer to tomorrow at the
    specified hour and at the current minute.

    :param hour: the hour for tomorrow, in the format *0-23* (defaults to
        ``None``)
    :type hour: int
    :param minute: the minute for tomorrow, in the format *0-59* (defaults to
        ``None``)
    :type minute: int
    :returns: a ``datetime.datetime`` object
    :raises: *ValueError* when hour or minute have bad values

    """
    now = datetime.now(timezone.utc)
    if hour is None:
        hour = now.hour
    if minute is None:
        minute = now.minute
    tomorrow_date = now.date() + timedelta(days=1)
    return datetime(tomorrow_date.year, tomorrow_date.month, tomorrow_date.day,
                    hour, minute, 0, 0, timezone.utc)


def yesterday(hour=None, minute=None):
    """
    Gives the ``datetime.datetime`` object corresponding to yesterday. The
    default value for optional parameters is the current value of hour and
    minute. I.e: when called without specifying values for parameters, the
    resulting object will refer to the time = now - 24 hours; when called with
    only hour specified, the resulting object will refer to yesterday at the
    specified hour and at the current minute.

    :param hour: the hour for yesterday, in the format *0-23* (defaults to
        ``None``)
    :type hour: int
    :param minute: the minute for yesterday, in the format *0-59* (defaults to
        ``None``)
    :type minute: int
    :returns: a ``datetime.datetime`` object
    :raises: *ValueError* when hour or minute have bad values
    """
    now = datetime.now(timezone.utc)
    if hour is None:
        hour = now.hour
    if minute is None:
        minute = now.minute
    yesterday_date = now.date() + timedelta(days=-1)
    return datetime(yesterday_date.year, yesterday_date.month,
                    yesterday_date.day, hour, minute, 0, 0, timezone.utc)


def next_week(date=None):
    """
    Gives the ``datetime.datetime`` object corresponding to the next week
    from now or from the specified ``datetime.datetime`` object. A week
    corresponds to 7 days.

    :param date: the date you want a week to be added (if left ``None``,
        the current date and time will be used)
    :type date: ``datetime.datetime`` object
    :returns: a ``datetime.datetime`` object
    """
    return _timedelta_days(7, date)


def last_week(date=None):
    """
    Gives the ``datetime.datetime`` object corresponding to the last week
    before now or before the specified ``datetime.datetime`` object. A week
    corresponds to 7 days.

    :param date: the date you want a week to be subtracted from (if left
       ``None``, the current date and time will be used)
    :type date: ``datetime.datetime`` object
    :returns: a ``datetime.datetime`` object
    """
    return _timedelta_days(-7, date)


def last_month(date=None):
    """
    Gives the ``datetime.datetime`` object corresponding to the last month
    before now or before the specified ``datetime.datetime`` object. A month
    corresponds to 30 days.

    :param date: the date you want a month to be subtracted from (if left
       ``None``, the current date and time will be used)
    :type date: ``datetime.datetime`` object
    :returns: a ``datetime.datetime`` object
    """
    return _timedelta_months(-1, date)


def next_month(date=None):
    """
    Gives the ``datetime.datetime`` object corresponding to the next month
    after now or after the specified ``datetime.datetime`` object. A month
    corresponds to 30 days.

    :param date: the date you want a month to be added to (if left
       ``None``, the current date and time will be used)
    :type date: ``datetime.datetime`` object
    :returns: a ``datetime.datetime`` object
    """
    return _timedelta_months(1, date)


def last_year(date=None):
    """
    Gives the ``datetime.datetime`` object corresponding to the last year
    before now or before the specified ``datetime.datetime`` object. A year
    corresponds to 365 days.

    :param date: the date you want a year to be subtracted from (if left
       ``None``, the current date and time will be used)
    :type date: ``datetime.datetime`` object
    :returns: a ``datetime.datetime`` object
    """
    return _timedelta_years(-1, date)


def next_year(date=None):
    """
    Gives the ``datetime.datetime`` object corresponding to the next year
    after now or after the specified ``datetime.datetime`` object. A month
    corresponds to 30 days.

    :param date: the date you want a year to be added to (if left
       ``None``, the current date and time will be used)
    :type date: ``datetime.datetime`` object
    :returns: a ``datetime.datetime`` object
    """
    return _timedelta_years(1, date)


def _timedelta_hours(offset, date=None):
    if date is None:
        return datetime.now(timezone.utc) + timedelta(hours=offset)
    assert isinstance(date, datetime), __name__ + \
        ": 'date' must be a datetime.datetime object"
    return date + timedelta(hours=offset)


def _timedelta_days(offset, date=None):
    if date is None:
        return datetime.now(timezone.utc) + timedelta(days=offset)
    assert isinstance(date, datetime), __name__ + \
        ": 'date' must be a datetime.datetime object"
    return date + timedelta(days=offset)


def _timedelta_months(offset, date=None):
    if date is None:
        return datetime.now(timezone.utc) + timedelta(days=offset * 30)
    assert isinstance(date, datetime), __name__ + \
        ": 'date' must be a datetime.datetime object"
    return date + timedelta(days=offset * 30)


def _timedelta_years(offset, date=None):
    if date is None:
        return datetime.now(timezone.utc) + timedelta(days=offset * 365)
    assert isinstance(date, datetime), __name__ + \
        ": 'date' must be a datetime.datetime object"
    return date + timedelta(days=offset * 365)


def millis_offset_between_epochs(reference_epoch, target_epoch):
    """
    Calculates the signed milliseconds delta between the reference unix epoch and the provided target unix epoch
    :param reference_epoch: the unix epoch that the millis offset has to be calculated with respect to
    :type reference_epoch: int
    :param target_epoch: the unix epoch for which the millis offset has to be calculated
    :type target_epoch: int
    :return: int
    """
    assert isinstance(reference_epoch, int)
    assert isinstance(target_epoch, int)
    return (target_epoch - reference_epoch)*1000



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pyowm.commons.exceptions
from pyowm.alertapi30.alert import Alert
from pyowm.alertapi30.condition import Condition
from pyowm.alertapi30.enums import AlertChannelsEnum
from pyowm.utils import formatting
from pyowm.utils.geo import GeometryBuilder


class Trigger:
    """
    Object representing a the check if a set of weather conditions are met on a given geographical area: each condition
    is a rule on the value of a given weather parameter (eg. humidity, temperature, etc). Whenever a condition from a
    `Trigger` is met, the OWM API crates an alert and binds it to the the `Trigger`.
    A `Trigger` is the local proxy for the corresponding entry on the OWM API, therefore it can get ouf of sync as
    time goes by and conditions are met: it's up to you to "refresh" the local trigger by using a
    `pyowm.utils.alertapi30.AlertManager` instance.
    :param start_after_millis: how many milliseconds after the trigger creation the trigger begins to be checked
    :type start_after_millis: int
    :param end_after_millis: how many milliseconds after the trigger creation the trigger ends to be checked
    :type end_after_millis: int
    :param alerts: the `Alert` objects representing the alerts that have been fired for this `Trigger` so far. Defaults
    to `None`
    :type alerts: list of `pyowm.utils.alertapi30.Alert` instances
    :param conditions: the `Condition` objects representing the set of checks to be done on weather variables
    :type conditions: list of `pyowm.utils.alertapi30.Condition` instances
    :param area: the geographic are over which conditions are checked: it can be composed by multiple geoJSON types
    :type area: list of geoJSON types
    :param alert_channels: the alert channels through which alerts originating from this `Trigger` can be consumed.
    Defaults to OWM API polling
    :type alert_channels: list of `pyowm.utils.alertapi30.AlertChannel` instances
    :param id: optional unique ID for this `Trigger` instance
    :type id: str
    :returns:  a *Trigger* instance
    :raises: *ValueError* when start or end epochs are `None` or when end precedes start or when conditions or area
    are empty collections

    """
    def __init__(self, start_after_millis, end_after_millis, conditions, area, alerts=None, alert_channels=None, id=None):
        assert start_after_millis is not None
        assert end_after_millis is not None
        assert isinstance(start_after_millis, int)
        assert isinstance(end_after_millis, int)
        if start_after_millis > end_after_millis:
            raise ValueError("Error: trigger start time must precede trigger end time")
        self.start_after_millis = start_after_millis
        self.end_after_millis = end_after_millis
        assert conditions is not None
        if len(conditions) == 0:
            raise ValueError('A trigger must contain at least one condition: you provided none')
        self.conditions = conditions
        assert area is not None
        if len(area) == 0:
            raise ValueError('The area for a trigger must contain at least one geoJSON type: you provided none')
        self.area = area
        if alerts is None or len(alerts) == 0:
            self.alerts = []
        else:
            self.alerts = alerts
        if alert_channels is None or len(alert_channels) == 0:
            self.alert_channels = [AlertChannelsEnum.OWM_API_POLLING]
        else:
            self.alert_channels = alert_channels
        self.id = id

    def get_alerts(self):
        """
        Returns all of the alerts for this `Trigger`
        :return: a list of `Alert` objects
        """
        return self.alerts

    def get_alert(self, alert_id):
        """
        Returns the `Alert` of this `Trigger` having the specified ID
        :param alert_id: str, the ID of the alert
        :return: `Alert` instance
        """
        for alert in self.alerts:
            if alert.id == alert_id:
                return alert
        return None

    def get_alerts_since(self, timestamp):
        """
        Returns all the `Alert` objects of this `Trigger` that were fired since the specified timestamp.
        :param timestamp: time object representing the point in time since when alerts have to be fetched
        :type timestamp: int, ``datetime.datetime`` or ISO8601-formatted string
        :return: list of `Alert` instances
        """
        unix_timestamp = formatting.to_UNIXtime(timestamp)
        return [alert for alert in self.alerts if alert.last_update >= unix_timestamp]

    def get_alerts_on(self, weather_param):
        """
        Returns all the `Alert` objects of this `Trigger` that refer to the specified weather parameter (eg. 'temp',
        'pressure', etc.). The allowed weather params are the ones enumerated by class
        `pyowm.alertapi30.enums.WeatherParametersEnum`
        :param weather_param: str, values in `pyowm.alertapi30.enums.WeatherParametersEnum`
        :return: list of `Alert` instances
        """
        result = []
        for alert in self.alerts:
            for met_condition in alert.met_conditions:
                if met_condition['condition'].weather_param == weather_param:
                    result.append(alert)
                    break
        return result

    @classmethod
    def from_dict(cls, the_dict):
        if the_dict is None:
            raise pyowm.commons.exceptions.ParseAPIResponseError('Data is None')
        try:
            # trigger id
            trigger_id = the_dict.get('_id', None)

            # start timestamp
            start_dict = the_dict['time_period']['start']
            expr = start_dict['expression']
            if expr != 'after':
                raise ValueError('Invalid time expression: "%s" on start timestamp. Only: "after" is supported' % expr)
            start = start_dict['amount']

            # end timestamp
            end_dict = the_dict['time_period']['end']
            expr = end_dict['expression']
            if expr != 'after':
                raise ValueError('Invalid time expression: "%s" on end timestamp. Only: "after" is supported' % expr)
            end = end_dict['amount']

            # conditions
            conditions = [Condition.from_dict(c) for c in the_dict['conditions']]

            # alerts
            alerts_dict = the_dict['alerts']
            alerts = list()
            for key in alerts_dict:
                alert_id = key
                alert_data = alerts_dict[alert_id]
                alert_last_update = alert_data['last_update']
                alert_met_conds = []
                for c in alert_data['conditions']:
                    if isinstance(c['current_value'], int):
                        cv = c['current_value']
                    else:
                        cv = c['current_value']['min']
                    item = dict(current_value=cv, condition=Condition.from_dict(c['condition']))
                    alert_met_conds.append(item)
                alert_coords = alert_data['coordinates']
                alert = Alert(alert_id, trigger_id, alert_met_conds, alert_coords, last_update=alert_last_update)
                alerts.append(alert)

            # area
            area_list = the_dict['area']
            area = [GeometryBuilder.build(a_dict) for a_dict in area_list]

            # alert channels
            alert_channels = None  # defaulting

        except ValueError as e:
            raise pyowm.commons.exceptions.ParseAPIResponseError(
                'Impossible to parse JSON: %s' % e
            )

        except KeyError as e:
            raise pyowm.commons.exceptions.ParseAPIResponseError(
                'Impossible to parse JSON: %s' % e
            )

        return Trigger(start, end, conditions, area=area, alerts=alerts, alert_channels=alert_channels, id=trigger_id)

    def to_dict(self):
        return {
            "start_after_millis": self.start_after_millis,
            "end_after_millis": self.end_after_millis,
            "conditions": [c.to_dict() for c in self.conditions],
            "area": [g.to_dict() for g in self.area],
            "alerts": [alert.to_dict() for alert in self.alerts],
            "alert_channels": [ac.to_dict() for ac in self.alert_channels],
            "id": self.id
        }

    def __repr__(self):
        return "<%s.%s - id=%s, start_after_mills=%s, end_after_mills=%s, alerts=%s>" % (
                    __name__,
                    self.__class__.__name__,
                    self.id if self.id is not None else 'None',
                    self.start_after_millis,
                    self.end_after_millis,
                    str(len(self.alerts)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
from pyowm.commons.image import Image
from pyowm.utils.geo import Polygon


class Tile:

    """
    Wrapper class for an image tile
    :param x: horizontal tile number in OWM tile reference system
    :type x: int
    :param y: vertical tile number in OWM tile reference system
    :type y: int
    :param zoom: zoom level for the tile
    :type zoom: int
    :param map_layer: the name of the OWM map layer this tile belongs to
    :type map_layer: str
    :param image: raw image data
    :type image: `pyowm.commons.image.Image instance`
    """

    def __init__(self, x, y, zoom, map_layer, image):
        assert x >= 0, 'X tile coordinate cannot be negative'
        self.x = x
        assert y >= 0, 'Y tile coordinate cannot be negative'
        self.y = y
        assert zoom >= 0, 'Tile zoom level cannot be negative'
        self.zoom = zoom
        self.map_layer = map_layer
        assert isinstance(image, Image), 'The provided image is in invalid format'
        self.image = image

    def persist(self, path_to_file):
        """
        Saves the tile to disk on a file

        :param path_to_file: path to the target file
        :type path_to_file: str
        :return: `None`
        """
        self.image.persist(path_to_file)

    def bounding_polygon(self):
        """
        Returns the bounding box polygon for this tile

        :return: `pywom.utils.geo.Polygon` instance
        """
        lon_left, lat_bottom, lon_right, lat_top = Tile.tile_coords_to_bbox(self.x, self.y, self.zoom)
        return Polygon([[[lon_left, lat_top],
                       [lon_right, lat_top],
                       [lon_right, lat_bottom],
                       [lon_left, lat_bottom],
                       [lon_left, lat_top]]])

    @classmethod
    def tile_coords_for_point(cls, geopoint, zoom):
        """
        Returns the coordinates of the tile containing the specified geopoint at the specified zoom level

        :param geopoint: the input geopoint instance
        :type geopoint: `pywom.utils.geo.Point`
        :param zoom: zoom level
        :type zoom: int
        :return: a tuple (x, y) containing the tile-coordinates
        """
        return Tile.geoocoords_to_tile_coords(geopoint.lon, geopoint.lat, zoom)

    @classmethod
    def geoocoords_to_tile_coords(cls, lon, lat, zoom):
        """
        Calculates the tile numbers corresponding to the specified geocoordinates at the specified zoom level
        Coordinates shall be provided in degrees and using the Mercator Projection (http://en.wikipedia.org/wiki/Mercator_projection)

        :param lon: longitude
        :type lon: int or float
        :param lat: latitude
        :type lat: int or float
        :param zoom: zoom level
        :type zoom: int
        :return: a tuple (x, y) containing the tile-coordinates
        """
        n = 2.0 ** zoom
        x = int((lon + 180.0) / 360.0 * n)
        y = int((1.0 - math.log(math.tan(math.radians(lat)) + (1 / math.cos(math.radians(lat)))) / math.pi) / 2.0 * n)
        return x, y

    @classmethod
    def tile_coords_to_bbox(cls, x, y, zoom):
        """
        Calculates the lon/lat estrema of the bounding box corresponding to specific tile coordinates. Output coordinates
        are in degrees and in the Mercator Projection (http://en.wikipedia.org/wiki/Mercator_projection)

        :param x: the x tile coordinates
        :param y: the y tile coordinates
        :param zoom: the zoom level
        :return: tuple with (lon_left, lat_bottom, lon_right, lat_top)
        """
        def tile_to_geocoords(x, y, zoom):
            n = 2. ** zoom
            lon = x / n * 360. - 180.
            lat = math.degrees(math.atan(math.sinh(math.pi * (1 - 2 * y / n))))
            return lat, lon
        north_west_corner =  tile_to_geocoords(x, y, zoom)
        south_east_corner = tile_to_geocoords(x+1, y+1, zoom)
        return north_west_corner[1], south_east_corner[0], south_east_corner[1], north_west_corner[0]

    def __repr__(self):
        return "<%s.%s - x=%s, y=%s, zoom=%s, map_layer=%s>" % (
            __name__, self.__class__.__name__, self.x, self.y, self.zoom, self.map_layer)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from calendar import timegm
from datetime import datetime, timedelta, timezone, tzinfo

ZERO = timedelta(0)


class UTC(tzinfo):
    """UTC"""

    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return ZERO


def timeformat(timeobject, timeformat):
    """
    Formats the specified time object to the target format type.

    :param timeobject: the object conveying the time value
    :type timeobject: int, ``datetime.datetime`` or ISO8601-formatted
        string with pattern ``YYYY-MM-DD HH:MM:SS+00:00``
    :param timeformat: the target format for the time conversion. May be:
        '*unix*' (outputs an int UNIXtime), '*date*' (outputs a
        ``datetime.datetime`` object) or '*iso*' (outputs an ISO8601-formatted
        string with pattern ``YYYY-MM-DD HH:MM:SS+00:00``)
    :type timeformat: str
    :returns: the formatted time
    :raises: ValueError when unknown timeformat switches are provided or
        when negative time values are provided
    """
    if timeformat == "unix":
        return to_UNIXtime(timeobject)
    elif timeformat == "iso":
        return to_ISO8601(timeobject)
    elif timeformat == "date":
        return to_date(timeobject)
    else:
        raise ValueError("Invalid value for timeformat parameter")


def to_date(timeobject):
    """
    Returns the ``datetime.datetime`` object corresponding to the time value
    conveyed by the specified object, which can be either a UNIXtime, a
    ``datetime.datetime`` object or an ISO8601-formatted string in the format
    `YYYY-MM-DD HH:MM:SS+00:00``.

    :param timeobject: the object conveying the time value
    :type timeobject: int, ``datetime.datetime`` or ISO8601-formatted
        string
    :returns: a ``datetime.datetime`` object
    :raises: *TypeError* when bad argument types are provided, *ValueError*
        when negative UNIXtimes are provided
    """
    if isinstance(timeobject, int):
        if timeobject < 0:
            raise ValueError("The time value is a negative number")
        return datetime.fromtimestamp(timeobject, tz=timezone.utc)
    elif isinstance(timeobject, datetime):
        return timeobject.replace(microsecond=0)
    elif isinstance(timeobject, str):
        return datetime.fromisoformat(timeobject)
    else:
        raise TypeError('The time value must be expressed either by an int ' \
                         'UNIX time, a datetime.datetime object or an ' \
                         'ISO8601-formatted string')


def to_ISO8601(timeobject):
    """
    Returns the ISO8601-formatted string corresponding to the time value
    conveyed by the specified object, which can be either a UNIXtime, a
    ``datetime.datetime`` object or an ISO8601-formatted string in the format
    `YYYY-MM-DD HH:MM:SS+00:00``.

    :param timeobject: the object conveying the time value
    :type timeobject: int, ``datetime.datetime`` or ISO8601-formatted
        string
    :returns: an ISO8601-formatted string with pattern
        `YYYY-MM-DD HH:MM:SS+00:00``
    :raises: *TypeError* when bad argument types are provided, *ValueError*
        when negative UNIXtimes are provided
    """
    if isinstance(timeobject, int):
        if timeobject < 0:
            raise ValueError("The time value is a negative number")
        return datetime.fromtimestamp(timeobject, tz=timezone.utc).isoformat(' ', 'seconds')
    elif isinstance(timeobject, datetime):
        return timeobject.isoformat(' ', 'seconds')
    elif isinstance(timeobject, str):
        return timeobject
    else:
        raise TypeError('The time value must be expressed either by an int ' \
                         'UNIX time, a datetime.datetime object or an ' \
                         'ISO8601-formatted string')


def to_UNIXtime(timeobject):
    """
    Returns the UNIXtime corresponding to the time value conveyed by the
    specified object, which can be either a UNIXtime, a
    ``datetime.datetime`` object or an ISO8601-formatted string in the format
    `YYYY-MM-DD HH:MM:SS+00:00``.

    :param timeobject: the object conveying the time value
    :type timeobject: int, ``datetime.datetime`` or ISO8601-formatted
        string
    :returns: an int UNIXtime
    :raises: *TypeError* when bad argument types are provided, *ValueError*
        when negative UNIXtimes are provided
    """
    if isinstance(timeobject, int):
        if timeobject < 0:
            raise ValueError("The time value is a negative number")
        return timeobject
    elif isinstance(timeobject, datetime):
        return datetime_to_UNIXtime(timeobject)
    elif isinstance(timeobject, str):
        return ISO8601_to_UNIXtime(timeobject)
    else:
        raise TypeError('The time value must be expressed either by an int ' \
                         'UNIX time, a datetime.datetime object or an ' \
                         'ISO8601-formatted string')


def ISO8601_to_UNIXtime(iso):
    """
    Converts an ISO8601-formatted string in the format
    ``YYYY-MM-DD HH:MM:SS+00:00`` to the correspondent UNIXtime

    :param iso: the ISO8601-formatted string
    :type iso: string
    :returns: an int UNIXtime
    :raises: *TypeError* when bad argument types are provided, *ValueError*
        when the ISO8601 string is badly formatted

    """
    try:
        d = datetime.fromisoformat(iso)
    except ValueError:
        raise ValueError(__name__ + ": bad format for input ISO8601 string, ' \
                'should have been: YYYY-MM-DD HH:MM:SS+00:00")
    return datetime_to_UNIXtime(d)


def datetime_to_UNIXtime(date):
    """
    Converts a ``datetime.datetime`` object to its correspondent UNIXtime

    :param date: the ``datetime.datetime`` object
    :type date: ``datetime.datetime``
    :returns: an int UNIXtime
    :raises: *TypeError* when bad argument types are provided
    """
    return int(date.timestamp())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.commons.enums import ImageTypeEnum
from pyowm.commons.http_client import HttpClient
from pyowm.commons.image import Image
from pyowm.commons.tile import Tile
from pyowm.tiles.uris import ROOT_TILE_URL, NAMED_MAP_LAYER_URL


class TileManager:

    """
    A manager objects that reads OWM map layers tile images .

    :param API_key: the OWM Weather API key
    :type API_key: str
    :param map_layer: the layer for which you want tiles fetched. Allowed map layers are specified by
        the `pyowm.tiles.enum.MapLayerEnum` enumerator class.
    :type map_layer: str
    :param config: the configuration dictionary
    :type config: dict
    :returns: a *TileManager* instance
    :raises: *AssertionError* when no API Key or no map layer is provided, or map layer name is not a string

    """

    def __init__(self, API_key, map_layer, config):
        assert API_key is not None, 'You must provide a valid API Key'
        self.API_key = API_key
        assert map_layer is not None, 'You must provide a valid map layer name'
        assert isinstance(map_layer, str), 'Map layer name must be a string'
        self.map_layer = map_layer
        assert isinstance(config, dict)
        self.http_client = HttpClient(API_key, config, ROOT_TILE_URL, admits_subdomains=False)

    def get_tile(self, x, y, zoom):
        """
        Retrieves the tile having the specified coordinates and zoom level

        :param x: horizontal tile number in OWM tile reference system
        :type x: int
        :param y: vertical tile number in OWM tile reference system
        :type y: int
        :param zoom: zoom level for the tile
        :type zoom: int
        :returns: a `pyowm.tiles.Tile` instance

        """
        status, data = self.http_client.get_png(
            NAMED_MAP_LAYER_URL % self.map_layer + '/%s/%s/%s.png' % (zoom, x, y),
            params={'appid': self.API_key})
        img = Image(data, ImageTypeEnum.PNG)
        return Tile(x, y, zoom, self.map_layer, img)

    def __repr__(self):
        return "<%s.%s - layer_name=%s>" % (__name__, self.__class__.__name__, self.map_layer)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.commons.databoxes import ImageType, SubscriptionType


class SubscriptionTypeEnum:
    """
    Allowed OpenWeatherMap subscription types

    """
    FREE = SubscriptionType('free', 'api', False)
    STARTUP = SubscriptionType('startup', 'api', True)
    DEVELOPER = SubscriptionType('developer', 'api', True)
    PROFESSIONAL = SubscriptionType('professional', 'api', True)
    ENTERPRISE = SubscriptionType('enterprise', 'api', True)

    @classmethod
    def lookup_by_name(cls, name):
        for i in SubscriptionTypeEnum.items():
            if i.name == name:
                return i
        raise ValueError('Subscription type not allowed')

    @classmethod
    def items(cls):
        """
        All values for this enum
        :return: list of `pyowm.commons.enums.SubscriptionType`

        """
        return [
            cls.FREE,
            cls.STARTUP,
            cls.DEVELOPER,
            cls.PROFESSIONAL,
            cls.ENTERPRISE
        ]

    def __repr__(self):
        return "<%s.%s>" % (__name__, self.__class__.__name__)


class ImageTypeEnum:
    """
    Allowed image types on OWM APIs

    """
    PNG = ImageType('PNG', 'image/png')
    GEOTIFF = ImageType('GEOTIFF', 'image/tiff')

    @classmethod
    def lookup_by_mime_type(cls, mime_type):
        for i in ImageTypeEnum.items():
            if i.mime_type == mime_type:
                return i
        return None

    @classmethod
    def lookup_by_name(cls, name):
        for i in ImageTypeEnum.items():
            if i.name == name:
                return i
        return None

    @classmethod
    def items(cls):
        """
        All values for this enum
        :return: list of `pyowm.commons.enums.ImageType`

        """
        return [
            cls.PNG,
            cls.GEOTIFF
        ]

    def __repr__(self):
        return "<%s.%s>" % (__name__, self.__class__.__name__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class SubscriptionType:
    """
    Databox class representing a type of subscription to OpenWeatherMap web APIs

    :param name: the name of the subscription
    :type name: str
    :param subdomain: the root API subdomain associated to the subscription
    :type subdomain: str
    :param is_paid: tells if the subscription plan is paid
    :type is_paid: bool
    """
    def __init__(self, name, subdomain, is_paid):

        self.name = name
        self.subdomain = subdomain
        self.is_paid = is_paid

    def __repr__(self):
        return "<%s.%s - name=%s subdomain=%s paid=%s>" % (
            __name__, self.__class__.__name__, self.name, self.subdomain, self.is_paid)


class ImageType:
    """
    Databox class representing an image type

    :param name: the image type name
    :type name: str
    :param mime_type: the image type MIME type
    :type mime_type: str
    """
    def __init__(self, name, mime_type):

        self.name = name
        self.mime_type = mime_type

    def __repr__(self):
        return "<%s.%s - name=%s mime=%s>" % (
            __name__, self.__class__.__name__, self.name, self.mime_type)


class Satellite:
    """
    Databox class representing a satellite

    :param name: the satellite
    :type name: str
    :param symbol: the short name of the satellite
    :type symbol: str
    """
    def __init__(self, name, symbol):

        self.name = name
        self.symbol = symbol

    def __repr__(self):
        return "<%s.%s - name=%s symbol=%s>" % (
            __name__, self.__class__.__name__, self.name, self.symbol)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

AGRO_API_VERSION = (1, 0, 0)
AIRPOLLUTION_API_VERSION = (3, 0, 0)
ALERT_API_VERSION = (3, 0, 0)
GEOCODING_API_VERSION = (1, 0, 0)
STATIONS_API_VERSION = (3, 0, 0)
UVINDEX_API_VERSION = (3, 0, 0)
WEATHER_API_VERSION = (2, 5, 0)
LANGUAGES = [
    "af",
    "al",
    "ar",
    "az",
    "bg",
    "ca",
    "cz",
    "da",
    "de",
    "el",
    "en",
    "es",
    "eu",
    "fa",
    "fi",
    "fr",
    "gl",
    "he",
    "hi",
    "hr",
    "hu",
    "id",
    "it",
    "ja",
    "kr",
    "la",
    "lt",
    "mk",
    "nl",
    "no",
    "pl",
    "pt",
    "pt_br",
    "ro",
    "ru",
    "se",
    "sk",
    "sl",
    "sp",
    "sr",
    "sv",
    "th",
    "tr",
    "ua",
    "uk",
    "vi",
    "zh_cn",
    "zh_tw",
    "zu",
]


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.commons import exceptions
from pyowm.weatherapi30 import location
from pyowm.utils import formatting, timestamps


class SO2Index:
    """
    A class representing the Sulphur Dioxide Index observed in a certain location
    in the world. The index is made up of several measurements, each one at a
    different atmospheric pressure. The location is represented by the
    encapsulated *Location* object.

    :param reference_time: GMT UNIXtime telling when the SO2 data has been measured
    :type reference_time: int
    :param location: the *Location* relative to this SO2 observation
    :type location: *Location*
    :param interval: the time granularity of the SO2 observation
    :type interval: str
    :param so2_samples: the SO2 samples
    :type so2_samples: list of dicts
    :param reception_time: GMT UNIXtime telling when the SO2 observation has
        been received from the OWM Weather API
    :type reception_time: int
    :returns: an *SOIndex* instance
    :raises: *ValueError* when negative values are provided as reception time,
      SO2 samples are not provided in a list

    """

    def __init__(self, reference_time, location, interval, so2_samples, reception_time):
        if reference_time < 0:
            raise ValueError("'reference_time' must be greater than 0")
        self.ref_time = reference_time
        self.location = location
        self.interval = interval
        if not isinstance(so2_samples, list):
            raise ValueError("'so2_samples' must be a list")
        self.so2_samples = so2_samples
        if reception_time < 0:
            raise ValueError("'reception_time' must be greater than 0")
        self.rec_time = reception_time

    def reference_time(self, timeformat='unix'):
        """
        Returns the GMT time telling when the SO2 samples have been measured

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time
            '*iso*' for ISO8601-formatted string in the format ``YYYY-MM-DD HH:MM:SS+00:00``
            '*date* for ``datetime.datetime`` object instance
        :type timeformat: str
        :returns: an int or a str
        :raises: ValueError when negative values are provided

        """
        return formatting.timeformat(self.ref_time, timeformat)

    def reception_time(self, timeformat='unix'):
        """
        Returns the GMT time telling when the SO2 observation has been received
        from the OWM Weather API

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time
            '*iso*' for ISO8601-formatted string in the format ``YYYY-MM-DD HH:MM:SS+00:00``
            '*date* for ``datetime.datetime`` object instance
        :type timeformat: str
        :returns: an int or a str
        :raises: ValueError when negative values are provided

        """
        return formatting.timeformat(self.rec_time, timeformat)

    def is_forecast(self):
        """
        Tells if the current SO2 observation refers to the future with respect
        to the current date
        :return: bool
        """
        return timestamps.now(timeformat='unix') < \
               self.reference_time(timeformat='unix')

    @classmethod
    def from_dict(cls, the_dict):
        """
        Parses a *SO2Index* instance out of a data dictionary. Only certain properties of the data dictionary
        are used: if these properties are not found or cannot be parsed, an exception is issued.

        :param the_dict: the input dictionary
        :type the_dict: `dict`
        :returns: a *SO2Index* instance or ``None`` if no data is available
        :raises: *ParseAPIResponseError* if it is impossible to find or parse the data needed to build the result

        """
        if the_dict is None:
            raise exceptions.ParseAPIResponseError('Data is None')
        try:
            # -- reference time (strip away Z and T on ISO8601 format)
            t = the_dict['time'].replace('Z', '+00:00').replace('T', ' ')
            reference_time = formatting.ISO8601_to_UNIXtime(t)

            # -- reception time (now)
            reception_time = timestamps.now('unix')

            # -- location
            lon = float(the_dict['location']['longitude'])
            lat = float(the_dict['location']['latitude'])
            place = location.Location(None, lon, lat, None)

            # -- SO2 samples
            so2_samples = the_dict['data']

        except KeyError:
            raise exceptions.ParseAPIResponseError(
                      ''.join([__name__, ': impossible to parse COIndex']))

        return SO2Index(reference_time, place, None, so2_samples, reception_time)

    def to_dict(self):
        """Dumps object to a dictionary

        :returns: a `dict`

        """
        return {"reference_time": self.ref_time,
                "location": self.location.to_dict(),
                "interval": self.interval,
                "so2_samples": self.so2_samples,
                "reception_time": self.rec_time}

    def __repr__(self):
        return "<%s.%s - reference time=%s, reception time=%s, location=%s, " \
               "interval=%s>" % (
                    __name__,
                    self.__class__.__name__,
                    self.reference_time('iso'),
                    self.reception_time('iso'),
                    str(self.location),
                    self.interval)



from typing import Union, Optional, List

from pyowm.commons import exceptions
from pyowm.utils import geo
from pyowm.weatherapi30.weather import Weather
from pyowm.weatherapi30.national_weather_alert import NationalWeatherAlert


class OneCall:

    def __init__(self,
                 lat: Union[int, float],
                 lon: Union[int, float],
                 timezone: str,
                 current: Weather,
                 forecast_minutely: Optional[List[Weather]] = None,
                 forecast_hourly: Optional[List[Weather]] = None,
                 forecast_daily: Optional[List[Weather]] = None,
                 national_weather_alerts: Optional[list] = None
                 ) -> None:
        geo.assert_is_lat(lat)
        self.lat = lat

        geo.assert_is_lon(lon)
        self.lon = lon

        self.timezone = timezone

        if current is None:
            raise ValueError("'current' must be set")
        self.current = current
        self.forecast_minutely = forecast_minutely
        self.forecast_hourly = forecast_hourly
        self.forecast_daily = forecast_daily
        self.national_weather_alerts = national_weather_alerts

    def __repr__(self):
        return "<%s.%s - lat=%s, lon=%s, retrieval_time=%s>" % (
            __name__, self.__class__.__name__, self.lat, self.lon,
            self.current.reference_time() if self.current else None)

    def to_geopoint(self):
        """
        Returns the geoJSON compliant representation of the place for this One Call

        :returns: a ``pyowm.utils.geo.Point`` instance

        """
        if self.lon is None or self.lat is None:
            return None
        return geo.Point(self.lon, self.lat)


    @classmethod
    def from_dict(cls, the_dict: dict):
        """
        Parses a *OneCall* instance out of a data dictionary. Only certain properties of the data dictionary
        are used: if these properties are not found or cannot be parsed, an exception is issued.

        :param the_dict: the input dictionary
        :type the_dict: `dict`
        :returns: a *OneCall* instance or ``None`` if no data is available
        :raises: *ParseAPIResponseError* if it is impossible to find or parse the
            data needed to build the result, *APIResponseError* if the input dict embeds an HTTP status error

        """

        if the_dict is None:
            raise exceptions.ParseAPIResponseError('Data is None')

        # Check if server returned errors: this check overcomes the lack of use
        # of HTTP error status codes by the OWM API 3.0. This mechanism is
        # supposed to be deprecated as soon as the API fully adopts HTTP for
        # conveying errors to the clients
        if 'message' in the_dict and 'cod' in the_dict:
            if the_dict['cod'] == "404":
                return None
            elif the_dict['cod'] == "429":
                raise exceptions.APIResponseError("OWM API: error - Exceeded call limit", the_dict['cod'])
            else:
                raise exceptions.APIResponseError("OWM API: error - response payload", the_dict['cod'])

        try:
            current = Weather.from_dict(the_dict["current"])
            minutely = None
            if "minutely" in the_dict:
                minutely = [Weather.from_dict(item) for item in the_dict["minutely"]]
            hourly = None
            if "hourly" in the_dict:
                hourly = [Weather.from_dict(item) for item in the_dict["hourly"]]
            daily = None
            if "daily" in the_dict:
                daily = [Weather.from_dict(item) for item in the_dict["daily"]]

        except KeyError:
            raise exceptions.ParseAPIResponseError(f"{__name__}: impossible to read weather info from input data")

        alerts = None
        if 'alerts' in the_dict:
            alerts = [NationalWeatherAlert.from_dict(item) for item in the_dict['alerts']]

        return OneCall(
            lat=the_dict.get("lat", None),
            lon=the_dict.get("lon", None),
            timezone=the_dict.get("timezone", None),
            current=current,
            forecast_minutely=minutely,
            forecast_hourly=hourly,
            forecast_daily=daily,
            national_weather_alerts=alerts
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# deprecated API endpoints
ROOT_POLLUTION_API_URL = 'openweathermap.org/pollution/v1'
CO_INDEX_URL = 'co'
OZONE_URL = 'o3'
NO2_INDEX_URL = 'no2'
SO2_INDEX_URL = 'so2'


# current API endpoint
NEW_ROOT_POLLUTION_API_URL = 'openweathermap.org/data/2.5'
AIR_POLLUTION_URL = 'air_pollution'
AIR_POLLUTION_FORECAST_URL = 'air_pollution/forecast'
AIR_POLLUTION_HISTORY_URL = 'air_pollution/history'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import datetime as dt
from pyowm.commons import exceptions
from pyowm.utils import formatting


class Station:
    """
    A class representing a meteostation in Stations API.
    A reference about OWM stations can be found at:
    http://openweathermap.org/stations

    :param id: unique OWM identifier for the station
    :type id: str
    :param created_at: UTC timestamp marking the station registration.
    :type created_at: str in format %Y-%m-%dT%H:%M:%S.%fZ
    :param updated_at: UTC timestamp marking the last update to this station
    :type updated_at: str in format %Y-%m-%dT%H:%M:%S.%fZ
    :param external_id: user-given identifier for the station
    :type external_id: str
    :param name: user-given name for the station
    :type name: str
    :param lon: longitude of the station
    :type lon: float
    :param lat: latitude of the station
    :type lat: float
    :param alt: altitude of the station
    :type alt: float
    :param rank: station rank
    :type rank: int
    """

    def __init__(self, id, created_at, updated_at, external_id, name,
                 lon, lat, alt, rank):
        assert id is not None
        assert external_id is not None
        assert lon is not None
        assert lat is not None
        if lon < -180.0 or lon > 180.0:
            raise ValueError("'lon' value must be between -180 and 180")
        self._lon = float(lon)
        if lat < -90.0 or lat > 90.0:
            raise ValueError("'lat' value must be between -90 and 90")
        if alt is not None and alt < 0.0:
            raise ValueError("'alt' value must not be negative")
        self.id = id
        self.created_at = created_at
        if self.created_at is not None:
            padded_created_at = self._format_micros(created_at)
            t = dt.fromisoformat(padded_created_at.replace("Z", "+00:00"))
            self.created_at = formatting.timeformat(t, 'unix')
        self.updated_at = updated_at
        if self.updated_at is not None:
            padded_updated_at = self._format_micros(updated_at)
            t = dt.fromisoformat(padded_updated_at.replace("Z", "+00:00"))
            self.updated_at = formatting.timeformat(t, 'unix')
        self.external_id = external_id
        self.name = name
        self.lon = lon
        self.lat = lat
        self.alt = alt
        self.rank = rank

    def _format_micros(self, datestring):
        parts = datestring[:-1].split('.')
        if len(parts) == 1:
            if datestring.endswith('Z'):
                return datestring[:-1] + '.000000Z'
            else:
                return datestring + '.000000Z'
        else:
            micros = parts[-1][:6] if len(parts[-1]) > 6 else parts[-1]
            return '.'.join(
                parts[:-1] + ['{:06d}'.format(int(micros))]) + 'Z'

    def creation_time(self, timeformat='unix'):
        """Returns the UTC time of creation of this station

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time, '*iso*' for ISO8601-formatted
            string in the format ``YYYY-MM-DD HH:MM:SS+00`` or `date` for
            a ``datetime.datetime`` object
        :type timeformat: str
        :returns: an int or a str or a ``datetime.datetime`` object or None
        :raises: ValueError

        """
        if self.created_at is None:
            return None
        return formatting.timeformat(self.created_at, timeformat)

    def last_update_time(self, timeformat='unix'):
        """Returns the UTC time of the last update on this station's metadata

        :param timeformat: the format for the time value. May be:
            '*unix*' (default) for UNIX time, '*iso*' for ISO8601-formatted
            string in the format ``YYYY-MM-DD HH:MM:SS+00`` or `date` for
            a ``datetime.datetime`` object
        :type timeformat: str
        :returns: an int or a str or a ``datetime.datetime`` object or None
        :raises: ValueError

        """
        if self.updated_at is None:
            return None
        return formatting.timeformat(self.updated_at, timeformat)

    @classmethod
    def from_dict(cls, the_dict):
        """
        Parses a *Station* instance out of a data dictionary. Only certain properties of the data dictionary
        are used: if these properties are not found or cannot be parsed, an exception is issued.

        :param the_dict: the input dictionary
        :type the_dict: `dict`
        :returns: a *Station* instance or ``None`` if no data is available
        :raises: *ParseAPIResponseError* if it is impossible to find or parse the data needed to build the result

        """
        if the_dict is None:
            raise exceptions.ParseAPIResponseError('Data is None')
        id = the_dict.get('ID', None) or the_dict.get('id', None)
        external_id = the_dict.get('external_id', None)
        lon = the_dict.get('longitude', None)
        lat = the_dict.get('latitude', None)
        alt = the_dict.get('altitude', None)
        name = the_dict.get('name', None)
        rank = the_dict.get('rank', None)
        created_at = the_dict.get('created_at', None)
        updated_at = the_dict.get('updated_at', None)
        return Station(id, created_at, updated_at, external_id, name, lon, lat, alt, rank)

    def to_dict(self):
        """Dumps object to a dictionary

        :returns: a `dict`

        """
        return {
            'id': self.id,
            'external_id': self.external_id,
            'name': self.name,
            'created_at': formatting.to_ISO8601(self.created_at),
            'updated_at': formatting.to_ISO8601(self.updated_at),
            'latitude': self.lat,
            'longitude': self.lon,
            'altitude': self.alt if self.alt is not None else 'None',
            'rank': self.rank}

    def __repr__(self):
        return '<%s.%s - id=%s, external_id=%s, name=%s>' \
               % (__name__, self.__class__.__name__,
                  self.id, self.external_id, self.name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.commons.http_client import HttpClient
from pyowm.constants import UVINDEX_API_VERSION
from pyowm.utils import formatting, geo, timestamps
from pyowm.uvindexapi30 import uv_client, uvindex
from pyowm.uvindexapi30.uris import ROOT_UV_API_URL


class UVIndexManager:

    """
    A manager objects that provides a full interface to OWM UV Index API.

    :param API_key: the OWM UV Index API key
    :type API_key: str
    :param config: the configuration dictionary
    :type config: dict
    :returns: a *UVIndexManager* instance
    :raises: *AssertionError* when no API Key is provided

    """

    def __init__(self, API_key, config):
        assert API_key is not None, 'You must provide a valid API Key'
        self.API_key = API_key
        assert isinstance(config, dict)
        self.uv_client = uv_client.UltraVioletHttpClient(
            API_key, HttpClient(API_key, config, ROOT_UV_API_URL))

    def uvindex_api_version(self):
        return UVINDEX_API_VERSION

    def uvindex_around_coords(self, lat, lon):
        """
        Queries for Ultra Violet value sampled in the
        surroundings of the provided geocoordinates and in the specified time
        interval. A *UVIndex* object instance is returned, encapsulating a
        *Location* object and the UV intensity value.

        :param lat: the location's latitude, must be between -90.0 and 90.0
        :type lat: int/float
        :param lon: the location's longitude, must be between -180.0 and 180.0
        :type lon: int/float
        :return: a *UVIndex* instance or ``None`` if data is not available
        :raises: *ParseResponseException* when OWM UV Index API responses' data
            cannot be parsed, *APICallException* when OWM UV Index API can not be
            reached, *ValueError* for wrong input values
        """
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        params = {'lon': lon, 'lat': lat}
        json_data = self.uv_client.get_uvi(params)
        return uvindex.UVIndex.from_dict(json_data)

    def uvindex_forecast_around_coords(self, lat, lon):
        """
        Queries for forecast Ultra Violet values in the next 8
        days in the surroundings of the provided geocoordinates.

        :param lat: the location's latitude, must be between -90.0 and 90.0
        :type lat: int/float
        :param lon: the location's longitude, must be between -180.0 and 180.0
        :type lon: int/float
        :return: a list of *UVIndex* instances or empty list if data is not available
        :raises: *ParseResponseException* when OWM UV Index API responses' data
            cannot be parsed, *APICallException* when OWM UV Index API can not be
            reached, *ValueError* for wrong input values
        """
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        params = {'lon': lon, 'lat': lat}
        json_data = self.uv_client.get_uvi_forecast(params)
        return [uvindex.UVIndex.from_dict(item) for item in json_data]

    def uvindex_history_around_coords(self, lat, lon, start, end=None):
        """
        Queries for UV index historical values in the
        surroundings of the provided geocoordinates and in the specified
        time frame. If the end of the time frame is not provided, that is
        intended to be the current datetime.

        :param lat: the location's latitude, must be between -90.0 and 90.0
        :type lat: int/float
        :param lon: the location's longitude, must be between -180.0 and 180.0
        :type lon: int/float
        :param start: the object conveying the time value for the start query boundary
        :type start: int, ``datetime.datetime`` or ISO8601-formatted string
        :param end: the object conveying the time value for the end query
            boundary (defaults to ``None``, in which case the current datetime
            will be used)
        :type end: int, ``datetime.datetime`` or ISO8601-formatted string
        :return: a list of *UVIndex* instances or empty list if data is not available
        :raises: *ParseResponseException* when OWM UV Index API responses' data
            cannot be parsed, *APICallException* when OWM UV Index API can not be
            reached, *ValueError* for wrong input values
        """
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        assert start is not None
        start = formatting.timeformat(start, 'unix')
        if end is None:
            end = timestamps.now(timeformat='unix')
        else:
            end = formatting.timeformat(end, 'unix')
        params = {'lon': lon, 'lat': lat, 'start': start, 'end': end}
        json_data = self.uv_client.get_uvi_history(params)
        return [uvindex.UVIndex.from_dict(item) for item in json_data]

    def __repr__(self):
        return '<%s.%s>' % (__name__, self.__class__.__name__)


#!/usr/bin/env python
# -*- coding: utf-8 -*-


class PyOWMError(Exception):
    """Generic base class for PyOWM exceptions"""
    pass


class ConfigurationError(PyOWMError):
    """Generic base class for configuration related errors"""
    pass


class ConfigurationNotFoundError(ConfigurationError):
    """Raised when configuration source file is not available"""
    pass


class ConfigurationParseError(ConfigurationError):
    """Raised on failures in parsing configuration data"""
    pass


class APIRequestError(PyOWMError):
    """
    Error class that represents network/infrastructural failures when invoking OWM Weather API, in
    example due to network errors.
    """
    pass


class BadGatewayError(APIRequestError):
    """
    Error class that represents 502 errors - i.e when upstream backend
    cannot communicate with API gateways.
    """
    pass


class TimeoutError(APIRequestError):
    """
    Error class that represents response timeout conditions
    """
    pass


class InvalidSSLCertificateError(APIRequestError):
    """
    Error class that represents failure in verifying the SSL certificate provided
    by the OWM API
    """
    pass


class APIResponseError(PyOWMError):
    """
    Generic base class for exceptions representing HTTP error status codes in OWM Weather API
    responses
    """
    pass


class NotFoundError(APIResponseError):
    """
    Error class that represents the situation when an entity is not found.
    """
    pass


class UnauthorizedError(APIResponseError):
    """
    Error class that represents the situation when an entity cannot be retrieved
    due to user subscription insufficient capabilities.
    """
    pass


class ParseAPIResponseError(PyOWMError):
    """
    Error class that represents failures when parsing payload data in HTTP
    responses sent by the OWM Weather API.
    """
    pass



from pyowm.commons.http_client import HttpClient
from pyowm.commons.uris import ROOT_GEOCODING_API_URL, DIRECT_GEOCODING_URI, REVERSE_GEOCODING_URI
from pyowm.constants import GEOCODING_API_VERSION
from pyowm.utils import geo
from pyowm.weatherapi30.location import Location


class GeocodingManager:

    """
    A manager objects that provides a full interface to OWM Geocoding API.

    :param API_key: the OWM API key
    :type API_key: str
    :param config: the configuration dictionary
    :type config: dict
    :returns: an *GeocodingManager* instance
    :raises: *AssertionError* when no API Key is provided

    """

    def __init__(self, API_key, config):
        assert API_key is not None, 'You must provide a valid API Key'
        self.API_key = API_key
        assert isinstance(config, dict)
        self.http_client = HttpClient(API_key, config, ROOT_GEOCODING_API_URL)

    def geocoding_api_version(self):
        return GEOCODING_API_VERSION

    def geocode(self, toponym, country=None, state_code=None, limit=None):
        """
        Invokes the direct geocoding API endpoint

        :param toponym: the name of the location
        :type toponym: `str`
        :param country: the 2-chars ISO symbol of the country
        :type country: `str` or `None`
        :param state_code: the 2-chars ISO symbol of state (only useful in case the country is US)
        :type state_code: `str` or `None`
        :param limit: the max number of results to be returned in case of multiple matchings (no limits by default)
        :type limit: `int` or `None`
        :returns: a list of *Location* instances
        :raises: *AssertionError*, *ValueError*, *APIRequestError*

        """
        assert toponym, 'Toponym must be specified'
        if country is not None and len(country) != 2:
            raise ValueError("Country must be a 2-char string")
        if state_code is not None and len(state_code) != 2:
            raise ValueError("State Code must be a 2-char string")
        if limit is not None:
            assert isinstance(limit, int)
            assert limit > 0

        query = toponym
        if state_code is not None:
            query += ',' + state_code
        if country is not None:
            query += ',' + country

        params = {'q': query}

        if limit is not None:
            params['limit'] = limit

        _, json_data = self.http_client.get_json(DIRECT_GEOCODING_URI, params=params)
        return [Location.from_dict(item) for item in json_data]

    def reverse_geocode(self, lat, lon, limit=None):
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        if limit is not None:
            assert isinstance(limit, int)
            assert limit > 0

        params = {'lat': lat, 'lon': lon}
        if limit is not None:
            params['limit'] = limit

        _, json_data = self.http_client.get_json(REVERSE_GEOCODING_URI, params=params)
        return [Location.from_dict(item) for item in json_data]

    def __repr__(self):
        return '<%s.%s>' % (__name__, self.__class__.__name__)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyowm.airpollutionapi30 import airpollution_client, coindex, no2index, ozone, so2index, airstatus
from pyowm.airpollutionapi30.uris import ROOT_POLLUTION_API_URL, NEW_ROOT_POLLUTION_API_URL
from pyowm.commons.http_client import HttpClient
from pyowm.constants import AIRPOLLUTION_API_VERSION
from pyowm.utils import geo, decorators, formatting, timestamps


class AirPollutionManager:

    """
    A manager objects that provides a full interface to OWM Air Pollution API.

    :param API_key: the OWM AirPollution API key
    :type API_key: str
    :param config: the configuration dictionary
    :type config: dict
    :returns: an *AirPollutionManager* instance
    :raises: *AssertionError* when no API Key is provided

    """

    def __init__(self, API_key, config):
        assert API_key is not None, 'You must provide a valid API Key'
        self.API_key = API_key
        assert isinstance(config, dict)
        self.ap_client = airpollution_client.AirPollutionHttpClient(
            API_key,
            HttpClient(API_key, config, ROOT_POLLUTION_API_URL))
        self.new_ap_client = airpollution_client.AirPollutionHttpClient(
            API_key,
            HttpClient(API_key, config, NEW_ROOT_POLLUTION_API_URL))

    def airpollution_api_version(self):
        return AIRPOLLUTION_API_VERSION

    @decorators.deprecated('removed', '4', 'coindex_around_coords')
    def coindex_around_coords(self, lat, lon, start=None, interval=None):
        """
        Queries the OWM AirPollution API for Carbon Monoxide values sampled in the
        surroundings of the provided geocoordinates and in the specified time
        interval.
        A *COIndex* object instance is returned, encapsulating a
        *Location* object and the list of CO samples
        If `start` is not provided, the latest available CO samples are
        retrieved
        If `start` is provided but `interval` is not, then `interval` defaults
        to the maximum extent, which is: `year`

        :param lat: the location's latitude, must be between -90.0 and 90.0
        :type lat: int/float
        :param lon: the location's longitude, must be between -180.0 and 180.0
        :type lon: int/float
        :param start: the object conveying the start value of the search time
            window start (defaults to ``None``). If not provided, the latest
            available CO samples value are retrieved
        :type start: int, ``datetime.datetime`` or ISO8601-formatted
            string
        :param interval: the length of the search time window starting at
           `start` (defaults to ``None``). If not provided, 'year' is used
        :type interval: str among: 'minute', 'hour', 'day', 'month, 'year'
        :return: a *COIndex* instance or ``None`` if data is not available
        :raises: *ParseResponseException* when OWM AirPollution API responses' data
            cannot be parsed, *APICallException* when OWM AirPollution API can not be
            reached, *ValueError* for wrong input values
        """
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        params = {'lon': lon, 'lat': lat, 'start': start, 'interval': interval}
        json_data = self.ap_client.get_coi(params)
        coi = coindex.COIndex.from_dict(json_data)
        if interval is None:
            interval = 'year'
        coi.interval = interval
        return coi

    @decorators.deprecated('removed', '4', 'ozone_around_coords')
    def ozone_around_coords(self, lat, lon, start=None, interval=None):
        """
        Queries the OWM AirPollution API for Ozone (O3) value in Dobson Units sampled in
        the surroundings of the provided geocoordinates and in the specified
        time interval. An *Ozone* object instance is returned, encapsulating a
        *Location* object and the UV intensity value.
        If `start` is not provided, the latest available ozone value is
        retrieved.
        If `start` is provided but `interval` is not, then `interval` defaults
        to the maximum extent, which is: `year`

        :param lat: the location's latitude, must be between -90.0 and 90.0
        :type lat: int/float
        :param lon: the location's longitude, must be between -180.0 and 180.0
        :type lon: int/float
        :param start: the object conveying the start value of the search time
            window start (defaults to ``None``). If not provided, the latest
            available Ozone value is retrieved
        :type start: int, ``datetime.datetime`` or ISO8601-formatted
            string
        :param interval: the length of the search time window starting at
           `start` (defaults to ``None``). If not provided, 'year' is used
        :type interval: str among: 'minute', 'hour', 'day', 'month, 'year'
        :return: an *Ozone* instance or ``None`` if data is not available
        :raises: *ParseResponseException* when OWM AirPollution API responses' data
            cannot be parsed, *APICallException* when OWM AirPollution API can not be
            reached, *ValueError* for wrong input values
        """
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        params = {'lon': lon, 'lat': lat, 'start': start, 'interval': interval}
        json_data = self.ap_client.get_o3(params)
        oz = ozone.Ozone.from_dict(json_data)
        if interval is None:
            interval = 'year'
        oz.interval = interval
        return oz

    @decorators.deprecated('removed', '4', 'no2index_around_coords')
    def no2index_around_coords(self, lat, lon, start=None, interval=None):
        """
        Queries the OWM AirPollution API for Nitrogen Dioxide values sampled in the
        surroundings of the provided geocoordinates and in the specified time
        interval.
        A *NO2Index* object instance is returned, encapsulating a
        *Location* object and the list of NO2 samples
        If `start` is not provided, the latest available NO2 samples are
        retrieved
        If `start` is provided but `interval` is not, then `interval` defaults
        to the maximum extent, which is: `year`

        :param lat: the location's latitude, must be between -90.0 and 90.0
        :type lat: int/float
        :param lon: the location's longitude, must be between -180.0 and 180.0
        :type lon: int/float
        :param start: the object conveying the start value of the search time
            window start (defaults to ``None``). If not provided, the latest
            available NO2 samples value are retrieved
        :type start: int, ``datetime.datetime`` or ISO8601-formatted
            string
        :param interval: the length of the search time window starting at
           `start` (defaults to ``None``). If not provided, 'year' is used
        :type interval: str among: 'minute', 'hour', 'day', 'month, 'year'
        :return: a *NO2Index* instance or ``None`` if data is not available
        :raises: *ParseResponseException* when OWM AirPollution API responses' data
            cannot be parsed, *APICallException* when OWM AirPollution API can not be
            reached, *ValueError* for wrong input values
        """
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        params = {'lon': lon, 'lat': lat, 'start': start, 'interval': interval}
        json_data = self.ap_client.get_no2(params)
        no2 = no2index.NO2Index.from_dict(json_data)
        if interval is None:
            interval = 'year'
        no2.interval = interval
        return no2

    @decorators.deprecated('removed', '4', 'so2index_around_coords')
    def so2index_around_coords(self, lat, lon, start=None, interval=None):
        """
        Queries the OWM AirPollution API for Sulphur Dioxide values sampled in the
        surroundings of the provided geocoordinates and in the specified time
        interval.
        A *SO2Index* object instance is returned, encapsulating a
        *Location* object and the list of SO2 samples
        If `start` is not provided, the latest available SO2 samples are
        retrieved
        If `start` is provided but `interval` is not, then `interval` defaults
        to the maximum extent, which is: `year`

        :param lat: the location's latitude, must be between -90.0 and 90.0
        :type lat: int/float
        :param lon: the location's longitude, must be between -180.0 and 180.0
        :type lon: int/float
        :param start: the object conveying the start value of the search time
            window start (defaults to ``None``). If not provided, the latest
            available SO2 samples value are retrieved
        :type start: int, ``datetime.datetime`` or ISO8601-formatted
            string
        :param interval: the length of the search time window starting at
           `start` (defaults to ``None``). If not provided, 'year' is used
        :type interval: str among: 'minute', 'hour', 'day', 'month, 'year'
        :return: a *SO2Index* instance or ``None`` if data is not available
        :raises: *ParseResponseException* when OWM AirPollution API responses' data
            cannot be parsed, *APICallException* when OWM AirPollution API can not be
            reached, *ValueError* for wrong input values
        """
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        params = {'lon': lon, 'lat': lat, 'start': start, 'interval': interval}
        json_data = self.ap_client.get_so2(params)
        so2 = so2index.SO2Index.from_dict(json_data)
        if interval is None:
            interval = 'year'
        so2.interval = interval
        return so2

    def air_quality_at_coords(self, lat, lon):
        """
        Queries the OWM AirPollution API for available air quality indicators around the specified coordinates.

        :param lat: the location's latitude, must be between -90.0 and 90.0
        :type lat: int/float
        :param lon: the location's longitude, must be between -180.0 and 180.0
        :type lon: int/float
        :return: a *AirStatus* instance or ``None`` if data is not available
        :raises: *ParseResponseException* when OWM AirPollution API responses' data
            cannot be parsed, *APICallException* when OWM AirPollution API can not be
            reached, *ValueError* for wrong input values
        """
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        params = {'lon': lon, 'lat': lat}
        json_data = self.new_ap_client.get_air_pollution(params)
        try:
            return airstatus.AirStatus.from_dict(json_data)
        except:
            return None

    def air_quality_forecast_at_coords(self, lat, lon):
        """
        Queries the OWM AirPollution API for available forecasted air quality indicators around the specified coordinates.

        :param lat: the location's latitude, must be between -90.0 and 90.0
        :type lat: int/float
        :param lon: the location's longitude, must be between -180.0 and 180.0
        :type lon: int/float
        :return: a `list` of *AirStatus* instances or an empty `list` if data is not available
        :raises: *ParseResponseException* when OWM AirPollution API responses' data
            cannot be parsed, *APICallException* when OWM AirPollution API can not be
            reached, *ValueError* for wrong input values
        """
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        params = {'lon': lon, 'lat': lat}
        json_data = self.new_ap_client.get_forecast_air_pollution(params)
        try:
            return airstatus.AirStatus.from_dict(json_data)
        except:
            return []

    def air_quality_history_at_coords(self, lat, lon, start, end=None):
        """
        Queries the OWM AirPollution API for available forecasted air quality indicators around the specified coordinates.

        :param lat: the location's latitude, must be between -90.0 and 90.0
        :type lat: int/float
        :param lon: the location's longitude, must be between -180.0 and 180.0
        :type lon: int/float
        :param start: the object conveying the start value of the search time window
        :type start: int, ``datetime.datetime`` or ISO8601-formatted string
        :param end: the object conveying the end value of the search time window. Values in the future will be clipped
           to the current timestamp. Defaults to the current UNIX timestamp.
        :type end: int, ``datetime.datetime`` or ISO8601-formatted string
        :return: a `list` of *AirStatus* instances or an empty `list` if data is not available
        :raises: *ParseResponseException* when OWM AirPollution API responses' data
            cannot be parsed, *APICallException* when OWM AirPollution API can not be
            reached, *ValueError* for wrong input values
        """
        geo.assert_is_lon(lon)
        geo.assert_is_lat(lat)
        now = timestamps.now(timeformat='unix')
        assert start is not None
        start = formatting.timeformat(start, 'unix')
        if end is None:
            end = now
        else:
            end = formatting.timeformat(end, 'unix')
            if end > now:
                end = now

        params = {'lon': lon, 'lat': lat, 'start': start, 'end': end}
        json_data = self.new_ap_client.get_historical_air_pollution(params)
        try:
            return airstatus.AirStatus.from_dict(json_data)
        except:
            return []

    def __repr__(self):
        return '<%s.%s>' % (__name__, self.__class__.__name__)
